from queue import Queue
from threading import Thread
from typing import Dict, Any

import rclpy

from depthai_sdk.integrations.ros.ros_base import RosBase
from depthai_sdk.logger import LOGGER


def ros_thread(queue: Queue):
    rclpy.init()
    node = rclpy.create_node('DepthAI_SDK')
    publishers = dict()

    while rclpy.ok():
        msgs: Dict[str, Any] = queue.get(block=True)
        for topic, msg in msgs.items():
            if topic not in publishers:
                publishers[topic] = node.create_publisher(type(msg), topic, 10)
                LOGGER.info(f'SDK started publishing ROS messages to {topic}')
            publishers[topic].publish(msg)
        rclpy.spin_once(node, timeout_sec=0.001)  # 1ms timeout


class Ros2Streaming(RosBase):
    queue: Queue

    def __init__(self):
        self.queue = Queue(30)
        self.process = Thread(target=ros_thread, args=(self.queue,))
        self.process.start()
        super().__init__()

    # def update(self): # By RosBase
    # def new_msg(self): # By RosBase

    def new_ros_msg(self, topic: str, ros_msg):
        self.queue.put({topic: ros_msg})



import numpy as np
import depthai as dai
from depthai_sdk.classes.packets import SemanticSegmentation


def decode(nndata: dai.NNData):
    # after squeeze the data.shape is 4,512, 896
    output = np.squeeze(nndata.getLayerFp16('L0317_ReWeight_SoftMax'))
    # Output is an array of 4x512x896 == 1835008 bytes, and has values from 0 to 1
    output = output.reshape((4, 512, 896))
    classes = np.asarray([0, 1, 2, 3], dtype=np.uint8)
    indices = np.argmax(output, axis=0)
    outputMap = np.take(classes, indices, axis=0)
    return SemanticSegmentation(nndata, mask=outputMap)




from datetime import datetime
from typing import Union, Callable

from depthai_sdk.classes import FramePacket
from depthai_sdk.oak_outputs.syncing import SequenceNumSync
from depthai_sdk.logger import LOGGER
from depthai_sdk.trigger_action.actions.abstract_action import Action
from depthai_sdk.trigger_action.triggers.abstract_trigger import Trigger

__all__ = ['TriggerAction']


class TriggerAction(SequenceNumSync):
    """
    TriggerAction class represents a single trigger-action pair.
    """

    def __init__(self, trigger: Trigger, action: Union[Action, Callable]):
        """
        Args:
            trigger: Trigger specifying the condition that activates the action.
            action: Action that gets activated by the trigger.
        """
        self.last_trigger_time = datetime.min
        self.trigger = trigger
        self.action = action
        if isinstance(action, Action) and action.inputs:
            SequenceNumSync.__init__(self, len(action.inputs))

    def new_packet_trigger(self, packet: FramePacket) -> None:
        """
        This method is called when a new packet is received from the trigger input stream.

        Args:
            packet: Packet received from the input stream.
        """
        if self.trigger.condition(packet):
            trigger_time = datetime.now()
            if trigger_time - self.last_trigger_time > self.trigger.cooldown:
                logging.debug(f'Triggered at {trigger_time}')
                self.last_trigger_time = trigger_time
                self.action.activate()

    def new_packet_action(self, packet: FramePacket) -> None:
        """
        This method is called when a new packet is received from the action input streams.
        Primary purpose of this method is to provide a way to keep a track of the packets.

        Args:
            packet: Packet received from the input stream.
        """
        synced = self.sync(
            packet.msg.getSequenceNum(),
            packet.name,
            packet
        )
        if synced:
            self.action.on_new_packets(synced)



from datetime import timedelta
from fractions import Fraction
from pathlib import Path

import av
import depthai as dai

from depthai_sdk.recorders.video_writers import BaseWriter
from depthai_sdk.recorders.video_writers.utils import create_writer_dir


class VideoWriter(BaseWriter):
    """
    Writes raw streams to file
    """

    def __init__(self, path: Path, name: str, lossless: bool = False):
        """
        Args:
            path: Path to save the output. Either a folder or a file.
            name: Name of the stream.
            lossless: If True, save the stream without compression.
        """

        super().__init__(path, name)

        self._lossless = lossless

        self._fourcc: str = None
        self._format: str = None
        self._start_ts: timedelta = None

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def create_file_for_buffer(self, subfolder: str, buf_name: str):
        if self._buffers[buf_name] is None:
            raise RuntimeError(f"Buffer {buf_name} is not enabled")

        if len(self._buffers[buf_name]) == 0:
            return None

        frame = self._buffers[buf_name][0]
        self.create_file(subfolder, frame)

    def create_file(self, subfolder: str, frame: dai.ImgFrame):
        if self._lossless or frame.getType() == dai.ImgFrame.Type.RAW16:
            extension = 'avi'
        else:
            extension = 'mp4'

        path_to_file = create_writer_dir(self.path / subfolder, self.name, extension)

        if not path_to_file.endswith('.' + extension):
            path_to_file = path_to_file[:-4] + '.' + extension

        self._create_file(path_to_file, frame)

    def _create_file(self, path_to_file: str, frame: dai.ImgFrame):
        options = {}
        if self._lossless:
            self._fourcc = 'rawvideo'
        elif frame.getType() == dai.ImgFrame.Type.RAW16:
            self._fourcc = 'ffv1'
            self._format = 'gray16le'
        else:  # Mono/Color, encode
            self._fourcc = 'h264'
            options['crf'] = '15'

        self._file = av.open(path_to_file, 'w')
        self._stream = self._file.add_stream(self._fourcc)
        self._stream.options = options
        self._stream.time_base = Fraction(1, 1000)
        self._stream.codec_context.width = frame.getWidth()
        self._stream.codec_context.height = frame.getHeight()

    def write(self, img_frame: dai.ImgFrame):
        if self._file is None:
            self.create_file(subfolder='', frame=img_frame)
        if self._start_ts is None:
            self._start_ts = img_frame.getTimestampDevice()

        if img_frame.getType() == dai.ImgFrame.Type.YUV420p:
            video_format = 'yuv420p'
        elif img_frame.getType() == dai.ImgFrame.Type.NV12:
            video_format = 'nv12'
        elif img_frame.getType() in [dai.ImgFrame.Type.RAW8, dai.ImgFrame.Type.GRAY8]:
            video_format = 'gray'
        elif img_frame.getType() == dai.ImgFrame.Type.RAW16:
            video_format = 'gray16le'
        else:
            raise ValueError(f'Unsupported frame type: {img_frame.getType()}')

        video_frame = av.VideoFrame.from_ndarray(img_frame.getFrame(), format=video_format)

        ts = int((img_frame.getTimestampDevice() - self._start_ts).total_seconds() * 1e3)  # To milliseconds
        video_frame.pts = ts + 1

        for packet in self._stream.encode(video_frame):
            self._file.mux(packet)

    def close(self) -> None:
        """
        Close the file if it is open.
        """
        if self._file:
            # Flush stream
            for packet in self._stream.encode():
                self._file.mux(packet)

            # Close output file
            self._file.close()



from abc import ABC
from datetime import timedelta
from typing import Union, Callable

from depthai_sdk.components import Component

__all__ = ['Trigger']


class Trigger(ABC):
    """
    Base trigger represents a single trigger that can activate an action.
    """

    def __init__(self, input: Union[Component, Callable], condition: Callable, cooldown: Union[timedelta, int]):
        """
        Args:
            input: Input component or output of a component that will be used as input for the trigger.
            condition: Condition that will be used to check if the trigger should be activated.
            cooldown: Cooldown time in seconds. If the trigger is activated, it will be ignored for the next `cooldown` seconds.
        """
        if isinstance(input, Component):
            input = input.out.main

        if isinstance(cooldown, timedelta):
            cooldown = cooldown.total_seconds()

        if cooldown >= 0:
            self.cooldown = timedelta(seconds=cooldown)
        else:
            raise ValueError("Cooldown time must be a non-negative integer or "
                             "a timedelta object representing non-negative time difference")

        self.input = input
        self.condition = condition
        self.cooldown = timedelta(seconds=cooldown)



import depthai as dai
from itertools import cycle
import logging

logger = logging.getLogger(__name__)

LIMITS = {
    "confidence_threshold": (0, 255),
    "bilateral_sigma": (0, 255),
    "range": (0, 65535),
    "lrc_threshold": (0, 10),
    "dot_projector": (0, 1200),
    "illumination_led": (0, 1500),
}


def clamp(value, min_value, max_value):
    return max(min(value, max_value), min_value)


class ToFControl:
    def __init__(self, device: dai.Device):
        self.queue = None
        ctrl = dai.StereoDepthConfig()
        self.raw_cfg = ctrl.get()

    def set_input_queue(self, queue: dai.DataInputQueue):
        self.queue = queue

    def send_controls(self, tof_control: dai.RawToFConfig):
        """
        Send controls to the ToF node.
        """
        if self.queue is None:
            logger.error("Cannot send controls when replaying.")
            return

        logger.info(f"Sending controls to ToF node: {tof_control}")
        self.queue.send(tof_control)



from typing import Dict

from depthai_sdk.classes.enum import ResizeMode
from depthai_sdk.components.camera_control import CameraControl
from depthai_sdk.components.camera_helper import *
from depthai_sdk.components.component import Component, ComponentOutput
from depthai_sdk.components.parser import parse_resolution, parse_encode, encoder_profile_to_fourcc
from depthai_sdk.logger import LOGGER
from depthai_sdk.oak_outputs.xout.xout_base import XoutBase, StreamXout, ReplayStream
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames
from depthai_sdk.replay import Replay
from depthai_sdk.types import Resolution


class CameraComponent(Component):
    def __init__(self,
                 device: dai.Device,
                 pipeline: dai.Pipeline,
                 source: dai.CameraBoardSocket,
                 resolution: Optional[Union[
                     str, dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution
                 ]] = None,
                 fps: Optional[float] = None,
                 encode: Union[None, str, bool, dai.VideoEncoderProperties.Profile] = None,
                 sensor_type: Optional[dai.CameraSensorType] = None,
                 rotation: Optional[int] = None,
                 replay: Optional[Replay] = None,
                 name: Optional[str] = None,
                 args: Dict = None):
        """
        Creates Camera component. This abstracts ColorCamera/MonoCamera nodes and supports mocking the camera when
        recording is passed during OakCamera initialization. Mocking the camera will send frames from the host to the
        OAK device (via XLinkIn node).

        Args:
            device (dai.Device): OAK device
            pipeline (dai.Pipeline): OAK pipeline
            source (str or dai.CameraBoardSocket): Source of the camera. Either color/rgb/right/left
            resolution (optional): Camera resolution, eg. '800p' or '4k'
            fps (float, optional): Camera FPS
            encode: Encode streams before sending them to the host. Either True (use default), or mjpeg/h264/h265
            sensor_type: To force color/mono/tof camera
            rotation (int, optional): Rotate the camera by 90, 180, 270 degrees
            replay (Replay object): Replay object to use for mocking the camera
            name (str, optional): Name of the output stream
            args (Dict): Use user defined arguments when constructing the pipeline
        """
        super().__init__()
        # _replay should be set before .out, as it's used in .out
        self._replay: Optional[Replay] = replay
        self.out = self.Out(self)

        self._pipeline = pipeline
        self._device = device

        self.node: Optional[Union[dai.node.ColorCamera, dai.node.MonoCamera, dai.node.XLinkIn]] = None
        self.encoder: Optional[dai.node.VideoEncoder] = None

        self.stream: Optional[dai.Node.Output] = None  # Node output to be used as eg. an input into NN
        self.stream_size: Optional[Tuple[int, int]] = None  # Output size

        self._source = str(source)
        if self._source.startswith('CameraBoardSocket.'):
            self._source = self._source[len('CameraBoardSocket.'):]

        self._socket = source
        self._replay: Optional[Replay] = replay
        self._args: Dict = args

        self.name = name

        if rotation not in [None, 0, 90, 180, 270]:
            raise ValueError(f'Angle {rotation} not supported! Use 0, 90, 180, 270.')

        self._num_frames_pool = 10
        self._preview_num_frames_pool = 4

        if self.is_replay():
            stream_name = None
            for name, stream in self._replay.streams.items():
                if stream.get_socket() == self._socket:
                    stream_name = name
                    break
            if stream_name is None:
                raise Exception(f"{source} stream was not found in specified depthai-recording!")
            self._source = stream_name
            res = self._replay.getShape(self._source)
            # print('resolution', res)
            # resize = getResize(res, width=1200)
            # self._replay.setResizeColor(resize)
            stream = self._replay.streams[self._source]
            if stream.node is None:
                return  # Stream disabled

            self.node = stream.node
            # print('resize', resize)
            self.node.setMaxDataSize(res[0] * res[1] * 3)
            self.stream_size = res
            self.stream = self.node.out

            if rotation in [90, 180, 270]:
                rot_manip = self._create_rotation_manip(pipeline, rotation)
                self.node.out.link(rot_manip.inputImage)
                self.stream = rot_manip.out
                if rotation in [90, 270]:
                    self.stream_size = self.stream_size[::-1]
        # Livestreaming, not replay
        else:
            node_type: dai.node = None
            sensors = [f for f in device.getConnectedCameraFeatures() if f.socket == source]
            if len(sensors) == 0:
                raise Exception(f"No camera found on user-specified socket {source}")
            sensor = sensors[0]

            sensor_type = sensor_type or sensor.supportedTypes[0]
            if sensor_type == dai.CameraSensorType.COLOR:
                node_type = dai.node.ColorCamera
            elif sensor_type == dai.CameraSensorType.MONO:
                node_type = dai.node.MonoCamera
            else:
                raise Exception(f"{sensor} doesn't support either COLOR or MONO ")

            # Create the node, and set the socket
            self.node = pipeline.create(node_type)
            self.node.setBoardSocket(source)

        self._resolution_forced: bool = resolution is not None
        if resolution:
            self._set_resolution(resolution)
        if fps:
            self.set_fps(fps)

        # Default configuration for the nodes
        if isinstance(self.node, dai.node.ColorCamera):
            # DepthAI uses CHW (Planar) channel layout convention for NN inferencing
            self.node.setInterleaved(False)
            # DepthAI uses BGR color order convention for NN inferencing
            self.node.setColorOrder(dai.ColorCameraProperties.ColorOrder.BGR)
            self.node.setPreviewNumFramesPool(self._preview_num_frames_pool)

            if not self._resolution_forced:  # Find the closest resolution
                sensor = [f for f in device.getConnectedCameraFeatures() if f.socket == self.node.getBoardSocket()][0]
                sensor_type = dai.CameraSensorType.COLOR if dai.node.ColorCamera else dai.CameraSensorType.MONO
                targetWidthRes = 1300
                targetWidthIsp = targetWidthRes
                if self._args["defaultResolution"] == "min":
                    targetWidthRes = 0
                    targetWidthIsp = 1300  # Still keep the same target for the ISP
                elif self._args["defaultResolution"] == "max":
                    targetWidthRes = 1000000  # Some big number
                    targetWidthIsp = targetWidthRes
                res = getClosesResolution(sensor, sensor_type, width=targetWidthRes)
                self.node.setResolution(res)
                scale = getClosestIspScale(self.node.getIspSize(), width=targetWidthIsp,
                                           videoEncoder=(encode is not None))
                self.node.setIspScale(*scale)

            curr_size = self.node.getVideoSize()
            closest = getClosestVideoSize(*curr_size, videoEncoder=encode)
            self.node.setVideoSize(*closest)
            self.node.setVideoNumFramesPool(2)  # We will increase it later if we are streaming to host

            self.node.setPreviewSize(*self.node.getVideoSize())
            self.stream_size = self.node.getPreviewSize()
            self.stream = self.node.preview

        elif isinstance(self.node, dai.node.MonoCamera):
            self.stream_size = self.node.getResolutionSize()
            self.stream = self.node.out

        if rotation and not self.is_replay():
            if rotation == 180:
                self.node.setImageOrientation(dai.CameraImageOrientation.ROTATE_180_DEG)
            else:
                rot_manip = self._create_rotation_manip(pipeline, rotation)
                self.stream.link(rot_manip.inputImage)
                self.stream = rot_manip.out
                self.stream_size = self.stream_size[::-1]

        if encode:
            self.encoder = pipeline.createVideoEncoder()
            self._encoder_profile = parse_encode(encode)  # MJPEG by default
            self.encoder.setDefaultProfilePreset(self.get_fps(), self._encoder_profile)

            if self.is_replay():  # TODO - this might be not needed, we check for replay above and return
                # Create ImageManip to convert to NV12
                type_manip = pipeline.createImageManip()
                type_manip.setFrameType(dai.ImgFrame.Type.NV12)
                type_manip.setMaxOutputFrameSize(self.stream_size[0] * self.stream_size[1] * 3)

                self.stream.link(type_manip.inputImage)
                type_manip.out.link(self.encoder.input)
            elif self.is_mono():
                self.stream.link(self.encoder.input)
            elif self.is_color():
                self.node.video.link(self.encoder.input)
            else:
                raise ValueError('CameraComponent is neither Color, Mono, nor Replay!')

        if self._args:
            self._config_camera_args(self._args)

        # Runtime camera control
        self.control = CameraControl()
        self._control_xlink_in = None
        if not self.is_replay():
            self._control_xlink_in = pipeline.create(dai.node.XLinkIn)
            self._control_xlink_in.setStreamName(f"{self.node.id}_inputControl")
            self._control_xlink_in.out.link(self.node.inputControl)
            # CameraControl message doesn't use any additional data (only metadata)
            self._control_xlink_in.setMaxDataSize(1)

    def on_pipeline_started(self, device: dai.Device):
        if self._control_xlink_in is not None:
            queue = device.getInputQueue(self._control_xlink_in.getStreamName())
            self.control.set_input_queue(queue)

    def _create_rotation_manip(self, pipeline: dai.Pipeline, rotation: int):
        rot_manip = pipeline.createImageManip()
        rgb_rr = dai.RotatedRect()
        w, h = self.stream_size
        rgb_rr.center.x, rgb_rr.center.y = w // 2, h // 2
        rgb_rr.size.width, rgb_rr.size.height = (w, h) if rotation % 180 == 0 else (h, w)
        rgb_rr.angle = rotation
        rot_manip.initialConfig.setCropRotatedRect(rgb_rr, False)
        rot_manip.setMaxOutputFrameSize(w * h * 3)
        return rot_manip

    def config_camera(self,
                      # preview: Union[None, str, Tuple[int, int]] = None,
                      size: Union[None, Tuple[int, int], str] = None,
                      resize_mode: ResizeMode = ResizeMode.CROP,
                      fps: Optional[float] = None,
                      resolution: Optional[Resolution] = None
                      ) -> None:
        """
        Configure resolution, scale, FPS, etc.
        """
        # TODO: Should be mono/color camera agnostic. Also call this from __init__ if args is enabled
        if fps:
            self.set_fps(fps)
        if resolution:
            self._set_resolution(resolution)

        if size:
            from .parser import parse_size
            size_tuple = parse_size(size)

            if self._replay:
                self._replay.resize(self._source, size_tuple, resize_mode)
            elif self.is_color():
                self.node.setStillSize(*size_tuple)
                self.node.setVideoSize(*size_tuple)
                self.node.setPreviewSize(*size_tuple)
                if resize_mode != ResizeMode.CROP:
                    raise ValueError("Currently only ResizeMode.CROP is supported mode for specifying size!")
            else:
                # TODO: Use ImageManip to set mono frame size
                raise NotImplementedError("Not yet implemented")

    def _config_camera_args(self, args: Dict):
        if not isinstance(args, Dict):
            args = vars(args)  # Namespace -> Dict

        if self.is_color():
            self.config_camera(
                fps=args.get('rgbFps', None),
                resolution=args.get('rgbResolution', None),
            )
            self.config_color_camera(
                manual_focus=args.get('manualFocus', None),
                af_mode=args.get('afMode', None),
                awb_mode=args.get('awbMode', None),
                scene_mode=args.get('sceneMode', None),
                anti_banding_mode=args.get('antiBandingMode', None),
                effect_mode=args.get('effectMode', None),
                isp_scale=args.get('ispScale', None),
                sharpness=args.get('sharpness', None),
                luma_denoise=args.get('lumaDenoise', None),
                chroma_denoise=args.get('chromaDenoise', None),
            )
        elif self.is_mono():
            self.config_camera(
                fps=args.get('monoFps', None),
                resolution=args.get('monoResolution', None),
            )
        else:  # Replay
            self.config_camera(fps=args.get('fps', None))

    def control_with_nn(self, detection_component: 'NNComponent', auto_focus=True, auto_exposure=True, debug=False):
        """
        Control the camera AF/AE/AWB based on the object detection results.

        :param detection_component: NNComponent that will be used to control the camera
        :param auto_focus: Enable auto focus to the object
        :param auto_exposure: Enable auto exposure to the object
        """

        if not auto_focus and not auto_exposure:
            LOGGER.error('Attempted to control camera with NN, '
                          'but both Auto-Focus and Auto-Exposure were disabled! Attempt ignored.')
            return

        if 'NNComponent' not in str(type(detection_component)):
            raise ValueError('nn_component must be an instance of NNComponent!')
        if not detection_component.is_detector():
            raise ValueError('nn_component must be a object detection model (YOLO/MobileNetSSD based)!')

        from depthai_sdk.components.control_camera_with_nn import control_camera_with_nn

        control_camera_with_nn(
            pipeline=self._pipeline,
            camera_control=self.node.inputControl,
            nn_output=detection_component.node.out,
            resize_mode=detection_component._ar_resize_mode,
            resolution=self.node.getResolution(),
            nn_size=detection_component._size,
            af=auto_focus,
            ae=auto_exposure,
            debug=debug
        )

    def config_color_camera(self,
                            interleaved: Optional[bool] = None,
                            color_order: Union[None, dai.ColorCameraProperties.ColorOrder, str] = None,
                            # Cam control
                            manual_focus: Optional[int] = None,
                            af_mode: Optional[dai.CameraControl.AutoFocusMode] = None,
                            awb_mode: Optional[dai.CameraControl.AutoWhiteBalanceMode] = None,
                            scene_mode: Optional[dai.CameraControl.SceneMode] = None,
                            anti_banding_mode: Optional[dai.CameraControl.AntiBandingMode] = None,
                            effect_mode: Optional[dai.CameraControl.EffectMode] = None,
                            # IQ settings
                            isp_scale: Optional[Tuple[int, int]] = None,
                            sharpness: Optional[int] = None,
                            luma_denoise: Optional[int] = None,
                            chroma_denoise: Optional[int] = None,
                            ) -> None:
        if not self.is_color():
            LOGGER.info('Attempted to configure ColorCamera, '
                         'but this component doesn\'t have it. Config attempt ignored.')
            return

        if self.is_replay():
            LOGGER.info('Tried configuring ColorCamera, but replaying is enabled. Config attempt ignored.')
            return

        if interleaved is not None: self.node.setInterleaved(interleaved)
        if color_order:
            if isinstance(color_order, str):
                color_order = getattr(dai.ColorCameraProperties.ColorOrder, color_order.upper())
            self.node.setColorOrder(color_order)

        if manual_focus is not None: self.node.initialControl.setManualFocus(manual_focus)
        if af_mode: self.node.initialControl.setAutoFocusMode(af_mode)
        if awb_mode: self.node.initialControl.setAutoWhiteBalanceMode(awb_mode)
        if scene_mode: self.node.initialControl.setSceneMode(scene_mode)
        if anti_banding_mode: self.node.initialControl.setAntiBandingMode(anti_banding_mode)
        if effect_mode: self.node.initialControl.setEffectMode(effect_mode)
        # EQ settings
        if isp_scale:
            self._resolution_forced = True
            self.node.setIspScale(*isp_scale)

            self.node.setPreviewSize(*self.node.getIspSize())
            self.node.setVideoSize(*self.node.getIspSize())
            self.stream_size = self.node.getIspSize()
            self.stream = self.node.preview

        if sharpness is not None: self.node.initialControl.setSharpness(sharpness)
        if luma_denoise is not None: self.node.initialControl.setLumaDenoise(luma_denoise)
        if chroma_denoise is not None: self.node.initialControl.setChromaDenoise(chroma_denoise)

    def _set_resolution(self, resolution):
        if not self.is_replay():
            if isinstance(resolution, str) and resolution.lower() in ['max', 'maximum']:
                sensor = [f for f in self._device.getConnectedCameraFeatures() if f.socket == self._socket][0]
                resolution = get_max_resolution(type(self.node), sensor)
            else:
                resolution = parse_resolution(type(self.node), resolution)
            self.node.setResolution(resolution)
        # TODO: support potentially downscaling depthai-recording

    def is_replay(self) -> bool:
        return self._replay is not None

    def is_color(self) -> bool:
        return isinstance(self.node, dai.node.ColorCamera)

    def is_mono(self) -> bool:
        return isinstance(self.node, dai.node.MonoCamera)

    def get_fps(self) -> float:
        if self.is_replay():
            return self._replay.get_fps()
        else:
            return self.node.getFps()

    def set_fps(self, fps: float):
        if self.is_replay():
            self._replay.set_fps(fps)
        else:
            self.node.setFps(fps)

    def config_encoder_h26x(self,
                            rate_control_mode: Optional[dai.VideoEncoderProperties.RateControlMode] = None,
                            keyframe_freq: Optional[int] = None,
                            bitrate_kbps: Optional[int] = None,
                            num_b_frames: Optional[int] = None,
                            ):
        if self.encoder is None:
            raise Exception('Video encoder was not enabled!')
        if self._encoder_profile == dai.VideoEncoderProperties.Profile.MJPEG:
            raise Exception('Video encoder was set to MJPEG while trying to configure H26X attributes!')

        if rate_control_mode is not None:
            self.encoder.setRateControlMode(rate_control_mode)
        if keyframe_freq is not None:
            self.encoder.setKeyframeFrequency(keyframe_freq)
        if bitrate_kbps is not None:
            self.encoder.setBitrateKbps(bitrate_kbps)
        if num_b_frames is not None:
            self.encoder.setNumBFrames(num_b_frames)

    def config_encoder_mjpeg(self,
                             quality: Optional[int] = None,
                             lossless: bool = False
                             ):
        if self.encoder is None:
            raise Exception('Video encoder was not enabled!')
        if self._encoder_profile != dai.VideoEncoderProperties.Profile.MJPEG:
            raise Exception(
                f'Video encoder was set to {self._encoder_profile} while trying to configure MJPEG attributes!'
            )

        if quality is not None:
            self.encoder.setQuality(quality)
        if lossless is not None:
            self.encoder.setLossless(lossless)

    def get_stream_xout(self, fourcc: Optional[str] = None) -> StreamXout:
        if self.encoder is not None and fourcc is not None:
            return StreamXout(self.encoder.bitstream, name=self.name or self._source + '_bitstream')
        elif self.is_replay():
            return ReplayStream(self.name or self._source)
        elif self.is_mono():
            return StreamXout(self.stream, name=self.name or self._source + '_mono')
        else:  # ColorCamera
            self.node.setVideoNumFramesPool(self._num_frames_pool)
            self.node.setPreviewNumFramesPool(self._preview_num_frames_pool)
            # node.video instead of preview (self.stream) was used to reduce bandwidth
            # consumption by 2 (3bytes/pixel vs 1.5bytes/pixel)
            return StreamXout(self.node.video, name=self.name or self._source + '_video')

    def set_num_frames_pool(self, num_frames: int, preview_num_frames: Optional[int] = None):
        """
        Set the number of frames to be stored in the pool.

        :param num_frames: Number of frames to be stored in the pool.
        :param preview_num_frames: Number of frames to be stored in the pool for the preview stream.
        """
        if self.is_color():
            self._num_frames_pool = num_frames
            if preview_num_frames is not None:
                self._preview_num_frames_pool = preview_num_frames

    def get_fourcc(self) -> Optional[str]:
        if self.encoder is None:
            return None
        return encoder_profile_to_fourcc(self._encoder_profile)

    """
    Available outputs (to the host) of this component
    """

    class Out:
        class CameraOut(ComponentOutput):
            def __call__(self, device: dai.Device, fourcc: Optional[str] = None) -> XoutBase:
                return XoutFrames(self._comp.get_stream_xout(fourcc), fourcc).set_comp_out(self)

        class ReplayOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                return XoutFrames(ReplayStream(self._comp._source)).set_comp_out(self)

        class EncodedOut(CameraOut):
            def __call__(self, device: dai.Device) -> XoutBase:
                return super().__call__(device, fourcc=self._comp.get_fourcc())


        def __init__(self, camera_component: 'CameraComponent'):
            self.replay = self.ReplayOut(camera_component)
            self.camera = self.CameraOut(camera_component)
            self.encoded = self.EncodedOut(camera_component)

            self.main = self.replay if camera_component.is_replay() else self.camera



from typing import Any, List, Dict

import depthai as dai
from sensor_msgs.msg import CompressedImage, Image, PointCloud2, Imu

from depthai_sdk.integrations.ros.depthai2ros2 import DepthAi2Ros2
from depthai_sdk.oak_outputs.xout.xout_base import XoutBase
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames


class RosStream:
    datatype: dai.DatatypeEnum  # dai datatype of the stream
    xout: XoutBase  # From which Xout is the stream
    topic: str  # Topic name, eg. color/compressed
    ros_type: Any  # CompressedImage, Image, IMU, PointCloud2...


class RosBase:
    """
    Base class that is used by ros streaming component and mcap recorder.
    """

    def __init__(self):
        self.streams: Dict[str, RosStream]  # key = xlink stream name
        self.pointcloud: bool = False  # Convert depth -> pointcloud
        self.bridge: DepthAi2Ros2
        self.streams = dict()

    def update(self, device: dai.Device, xouts: List[XoutFrames]):
        self.bridge = DepthAi2Ros2(device)

        for xout in xouts:
            for stream in xout.xstreams():
                rs = RosStream()
                rs.datatype = stream.stream.possibleDatatypes[0].datatype
                name = xout.name.lower()

                if xout.is_depth() and not stream.name.endswith('depth'):
                    # Mono right frame for WLS, skip
                    continue

                if xout.is_mjpeg():
                    rs.topic = f'/{name}/compressed'
                    rs.ros_type = CompressedImage
                elif xout.is_depth() and self.pointcloud:
                    rs.topic = '/pointcloud/raw'
                    rs.ros_type = PointCloud2
                elif xout.is_imu():
                    rs.topic = '/imu'
                    rs.ros_type = Imu
                else:  # Non-encoded frames; rgb, mono, depth
                    rs.topic = f'/{name}/raw'
                    rs.ros_type = Image

                self.streams[stream.name] = rs

    def new_ros_msg(self, topic: str, ros_msg) -> None:
        raise NotImplementedError('Abstract function, override it!')

    def new_msg(self, name: str, dai_msg: dai.ADatatype):
        if name not in self.streams:  # Not relevant
            return

        # depthai msgs, name = xlink name
        stream = self.streams[name]

        msg = None
        if stream.ros_type == CompressedImage:
            dai_msg: dai.ImgFrame
            msg = self.bridge.CompressedImage(dai_msg)
        # elif stream.ros_type == PointCloud2:
        #     msg = self.bridge.PointCloud2(dai_msg)
        elif stream.ros_type == Imu:
            msg = self.bridge.Imu(dai_msg)
        elif stream.ros_type == Image:
            dai_msg: dai.ImgFrame
            msg = self.bridge.Image(dai_msg)

        self.new_ros_msg(stream.topic, msg)



import numpy as np
import cv2
import random
from typing import Tuple
import open3d as o3d
import json
from depthai_sdk.logger import LOGGER

N_POINTS_SAMPLED_PLANE = 3
MAX_ITER_PLANE = 300

class BoxEstimator:
    def __init__(self, median_window=3, calib_json_path: str = None, threshold=50):
        """
        Box estimator helper class. Currently it's applicable for scanning only one box at a time.

        Args:
            threshold (int, optional): Distance threshold for plane fitting. Defaults to 50 mm.
        """
        self.top_side_pcl = None

        self.ground_plane_eq = None
        self.threshold = threshold

        self.height = None
        self.width = None
        self.length = None

        self.bounding_box = None
        self.rotation_matrix = None
        self.translate_vector = None

        # Median filter
        self.median_window = median_window
        self.prev_dimensions = []

        self.pcd = o3d.geometry.PointCloud()
        self.plane_pcd = o3d.geometry.PointCloud()
        self.box_pcd = o3d.geometry.PointCloud()

        self.corners = None

        # Try to find plane_eq.json file (where user executed script)
        if calib_json_path is None:
            calib_json_path = 'plane_eq.json'

        try:
            with open(calib_json_path, 'r') as f:
                data = json.load(f)
                if 'eq' in data:
                    self.ground_plane_eq = np.array(data['eq'], dtype=np.float64)
                else:
                    raise ValueError("Invalid JSON file")
        except FileNotFoundError:
            LOGGER.info("No plane_eq.json file found. Please run calibrate() method to calibrate the ground plane.")

    def get_outliers(self, points) -> Tuple[np.ndarray, np.ndarray]:
        """
        Get outliers and inliers from a point cloud
        """
        A,B,C,D = self.ground_plane_eq
        # Calculate the denominator (constant for all points)
        denominator = np.sqrt(A**2 + B**2 + C**2)
        threshold = 45 # mm
        distances = np.abs(A * points[:, 0] + B * points[:, 1] + C * points[:, 2] + D) / denominator
        return (points[distances > self.threshold], points[distances <= self.threshold])

    def is_calibrated(self) -> bool:
        return self.ground_plane_eq is not None

    def calibrate(self, points: np.ndarray):
        """
        Calibrate the ground plane equation using a point cloud

        Args:
            points: Point cloud

        Returns:
            None
        """
        pcd = o3d.geometry.PointCloud()
        pcd.points = o3d.utility.Vector3dVector(points.reshape(-1, 3))
        pcd = pcd.voxel_down_sample(voxel_size=10)
        pcd = pcd.remove_statistical_outlier(30, 0.1)[0]

        # Get plane segmentation
        plane_eq, plane_inliers = pcd.segment_plane(self.threshold, N_POINTS_SAMPLED_PLANE, MAX_ITER_PLANE)
        inlier_points_num = len(plane_inliers)
        inline_percentage = inlier_points_num / len(points)
        if inline_percentage < 0.8:
            LOGGER.error("Not enough inliers found. Please try again.")
            return False

        self.ground_plane_eq = plane_eq
        with open('plane_eq.json', 'w') as f:
            # Convert np.array to list
            json.dump({'eq': self.ground_plane_eq.tolist()}, f)
        return True


    def get_plane_mesh(self, size=10, divisions=10) -> Tuple:
        """
        Create a mesh representation of a plane given its equation Ax + By + Cz + D = 0.

        Args:
            size: Size of the plane (default: 10)
            divisions: Number of divisions in each dimension (default: 10)

        Returns:
            positions: List of 3D points
            indices: List of indices for triangles
            normals: List of normal vectors for each vertex
        """
        # Normalize the normal vector
        A,B,C,D = self.ground_plane_eq
        normal = np.array([A, B, C])
        normal = normal / np.linalg.norm(normal)

        # Create a grid of points
        x = np.linspace(-size/2, size/2, divisions)
        y = np.linspace(-size/2, size/2, divisions)
        X, Y = np.meshgrid(x, y)

        # Calculate Z values
        if C != 0:
            Z = (-A*X - B*Y - D) / C
        elif B != 0:
            Z = (-A*X - C*Y - D) / B
            Y, Z = Z, Y
        else:
            Z = (-B*Y - C*Z - D) / A
            X, Z = Z, X

        # Create positions
        positions = np.stack((X, Y, Z), axis=-1).reshape(-1, 3).tolist()

        # Create indices
        indices = []
        for i in range(divisions - 1):
            for j in range(divisions - 1):
                square_start = i * divisions + j
                indices.extend([
                    square_start, square_start + 1, square_start + divisions,
                    square_start + 1, square_start + divisions + 1, square_start + divisions
                ])

        # Create normals
        normals = [normal.tolist() for _ in range(len(positions))]

        return positions, indices, normals

    def process_points(self, points_roi: np.ndarray) -> Tuple:
        self.pcd.points = o3d.utility.Vector3dVector(points_roi)
        self.pcd = self.pcd.voxel_down_sample(voxel_size=10)
        self.pcd = self.pcd.remove_statistical_outlier(30, 0.1)[0]

        self.box_pcl, self.plane_pcl = self.get_outliers(np.asarray(self.pcd.points))

        # Remove outliers
        self.plane_pcd = self.plane_pcd.remove_statistical_outlier(30, 0.1)[0]
        self.box_pcd = self.box_pcd.remove_statistical_outlier(30, 0.1)[0]


        if len(self.box_pcl) < 100:
            return None, None  # No box

        self.get_box_top(self.ground_plane_eq)
        dimensions = self.get_dimensions()
        self.prev_dimensions.append(dimensions)
        corners = self.get_3d_corners()
        return self._filtered_dimensions(), corners

    @property
    def box_pcl(self):
        return np.asarray(self.box_pcd.points)
    @box_pcl.setter
    def box_pcl(self, points):
        self.box_pcd.points = o3d.utility.Vector3dVector(points)
    @property
    def plane_pcl(self):
        return np.asarray(self.plane_pcd.points)
    @plane_pcl.setter
    def plane_pcl(self, points):
        self.plane_pcd.points = o3d.utility.Vector3dVector(points)

    def _filtered_dimensions(self):
        # If too many dimensions, remove the oldest one
        if len(self.prev_dimensions) > self.median_window:
            self.prev_dimensions.pop(0)

        # Get median of the dimensions
        length = np.median([d[0] for d in self.prev_dimensions])
        width = np.median([d[1] for d in self.prev_dimensions])
        height = np.median([d[2] for d in self.prev_dimensions])
        return length, width, height


    def create_rotation_matrix(self, vec_in, vec_target):
        v = np.cross(vec_in, vec_target)
        s = np.linalg.norm(v)
        c = np.dot(vec_in, vec_target)
        v_mat = np.array([[0, -v[2], v[1]],
                          [v[2], 0, -v[0]],
                          [-v[1], v[0], 0]])
        R = np.eye(3) + v_mat + (np.matmul(v_mat, v_mat) * (1 / (1 + c)))
        self.rotation_matrix = R
        return R

    def get_box_top(self, plane_eq):
        rot_matrix = self.create_rotation_matrix(plane_eq[0:3], [0, 0, 1])
        self.plane_pcl = self.rotate_points(self.plane_pcl, rot_matrix)
        avg_z = np.average(self.plane_pcl[:, 2])

        translate_vector = [0, 0, -avg_z]
        self.translate_vector = np.array(translate_vector)

        self.plane_pcl = self.translate_points(self.plane_pcl, translate_vector)
        self.box_pcl = self.rotate_points(self.box_pcl, rot_matrix)
        self.box_pcl = self.translate_points(self.box_pcl, translate_vector)

        top_plane_eq, top_plane_inliers = self.fit_plane_vec_constraint([0, 0, 1], self.box_pcl, 3, 30)

        top_plane = self.box_pcl[top_plane_inliers]
        # self.side_planes = self.box_pcl[np.setdiff1d(np.arange(len(self.box_pcl)), top_plane_inliers)]
        self.top_side_pcl = top_plane
        self.height = abs(top_plane_eq[3])
        return self.height

    def get_dimensions(self):
        upper_plane_points = self.top_side_pcl
        coordinates = np.c_[upper_plane_points[:, 0], upper_plane_points[:, 1]].astype('float32')
        rect = cv2.minAreaRect(coordinates)
        self.bounding_box = cv2.boxPoints(rect)
        self.width, self.length = rect[1][0], rect[1][1]

        return self.length, self.width, self.height

    def get_3d_lines(self, corners):
        lines = [
            [0, 1], [1, 2], [2, 3], [3, 0],  # Bottom square
            [4, 5], [5, 6], [6, 7], [7, 4],  # Top square
            [0, 4], [1, 5], [2, 6], [3, 7]   # Vertical lines
        ]
        line_segments = []
        for line in lines:
            line_segments.append(corners[line[0]])
            line_segments.append(corners[line[1]])
        return np.array(line_segments)

    def get_3d_corners(self):
        bounding_box = self.bounding_box
        points_floor = np.c_[bounding_box, np.zeros(4)]
        points_top = np.c_[bounding_box, -self.height * np.ones(4)]
        box_points = np.concatenate((points_floor, points_top))

        self.corners = box_points
        return box_points

    def inverse_corner_points(self):
        # Apply the inverse of the translation and rotation to get back to the original coordinates
        inverse_translation = -self.translate_vector
        inverse_rot_mat = np.linalg.inv(self.rotation_matrix)

        box_points = self.translate_points(self.corners, inverse_translation)
        box_points = self.rotate_points(box_points, inverse_rot_mat)
        return box_points

    def fit_plane_vec_constraint(self, norm_vec, pts, thresh=0.05, n_iterations=300):
        best_eq = []
        best_inliers = []

        n_points = pts.shape[0]
        for _ in range(n_iterations):
            id_sample = random.sample(range(0, n_points), 1)
            point = pts[id_sample]
            d = -np.sum(np.multiply(norm_vec, point))
            plane_eq = [*norm_vec, d]
            pt_id_inliers = self.get_plane_inliers(plane_eq, pts, thresh)
            if len(pt_id_inliers) > len(best_inliers):
                best_eq = plane_eq
                best_inliers = pt_id_inliers

        return best_eq, best_inliers

    def get_plane_inliers(self, plane_eq, pts, thresh=0.05):
        dist_pt = self.get_pts_distances_plane(plane_eq, pts)
        pt_id_inliers = np.where(np.abs(dist_pt) <= thresh)[0]
        return pt_id_inliers

    def get_pts_distances_plane(self, plane_eq, pts):
        dist_pt = (plane_eq[0] * pts[:, 0] + plane_eq[1] * pts[:, 1]
                   + plane_eq[2] * pts[:, 2] + plane_eq[3]) \
                  / np.sqrt(plane_eq[0] ** 2 + plane_eq[1] ** 2 + plane_eq[2] ** 2)
        return dist_pt

    def rotate_points(self, points, rotation_matrix):
        return np.dot(points, rotation_matrix.T)

    def translate_points(self, points, translate_vector):
        return points + translate_vector


import numpy as np


def _get_mesh(mapX: np.ndarray, mapY: np.ndarray):
    mesh_cell_size = 16
    mesh0 = []
    # Creates subsampled mesh which will be loaded on to device to undistort the image
    for y in range(mapX.shape[0] + 1):  # iterating over height of the image
        if y % mesh_cell_size == 0:
            row_left = []
            for x in range(mapX.shape[1] + 1):  # iterating over width of the image
                if x % mesh_cell_size == 0:
                    if y == mapX.shape[0] and x == mapX.shape[1]:
                        row_left.append(mapY[y - 1, x - 1])
                        row_left.append(mapX[y - 1, x - 1])
                    elif y == mapX.shape[0]:
                        row_left.append(mapY[y - 1, x])
                        row_left.append(mapX[y - 1, x])
                    elif x == mapX.shape[1]:
                        row_left.append(mapY[y, x - 1])
                        row_left.append(mapX[y, x - 1])
                    else:
                        row_left.append(mapY[y, x])
                        row_left.append(mapX[y, x])
            if (mapX.shape[1] % mesh_cell_size) % 2 != 0:
                row_left.append(0)
                row_left.append(0)

            mesh0.append(row_left)

    mesh0 = np.array(mesh0)
    # mesh = list(map(tuple, mesh0))
    return mesh0



from typing import Optional, Tuple, Union, Sequence, List

import depthai as dai
import numpy as np

from depthai_sdk.classes.enum import ResizeMode


class Point:
    """
    Used within the BoundingBox class when dealing with points.
    """

    def __init__(self, x: float, y: float):
        self.x = x
        self.y = y

    def __str__(self):
        return f"({self.x}, {self.y})"

    def to_tuple(self) -> Tuple[float, float]:
        return self.x, self.y

    def denormalize(self, frame_shape: Sequence) -> Tuple[int, int]:
        """
        Denormalize the point to pixel coordinates (0..frame width, 0..frame height)
        """
        return int(self.x * frame_shape[1]), int(self.y * frame_shape[0])


class BoundingBox:
    """
    This class helps with bounding box calculations. It can be used to calculate relative bounding boxes,
    map points from relative to absolute coordinates and vice versa, crop frames, etc.
    """

    def __init__(self, bbox: Union[None, List, np.ndarray, Tuple[float, float, float, float], dai.ImgDetection] = None):
        if isinstance(bbox, (Sequence, np.ndarray)):
            self.xmin, self.ymin, self.xmax, self.ymax = bbox
        elif isinstance(bbox, dai.ImgDetection):
            self.xmin, self.ymin, self.xmax, self.ymax = bbox.xmin, bbox.ymin, bbox.xmax, bbox.ymax
        else:
            self.xmin, self.ymin, self.xmax, self.ymax = 0.0, 0.0, 1.0, 1.0
        self.width, self.height = self.xmax - self.xmin, self.ymax - self.ymin

    def __str__(self):
        return f"({self.xmin}, {self.ymin}), ({self.xmax}, {self.ymax})"

    def clip(self, min_value=0.0, max_value=1.0) -> 'BoundingBox':
        """
        Clips the bounding box to the given range.
        """
        return BoundingBox([
            np.clip(self.xmin, min_value, max_value),
            np.clip(self.ymin, min_value, max_value),
            np.clip(self.xmax, min_value, max_value),
            np.clip(self.ymax, min_value, max_value),
        ])

    def top_left(self) -> Tuple[float, float]:
        """
        Get top-left corner of the bounding box.
        """
        return self.xmin, self.ymin

    def bottom_right(self) -> Tuple[float, float]:
        """
        Get bottom-right corner of the bounding box.
        """
        return self.xmax, self.ymax

    def to_tuple(self, frame_shape: Union[Sequence, None] = None) -> Tuple:
        """
        Get bounding box coordinates as a tuple (xmin, ymin, xmax, ymax).
        If frame_shape is passed, then it will return the coordinates in pixels (0..frame width, 0..frame height).
        """
        if frame_shape is not None:
            tl, br = self.denormalize(frame_shape)
            return *tl, *br
        return self.xmin, self.ymin, self.xmax, self.ymax

    def get_relative_bbox(self, bbox: 'BoundingBox') -> 'BoundingBox':
        """
        Calculate relative BoundingBox to the current BoundingBox.
        Example: First we run vehicle detection on a frame, crop out a frame only of the vehicle, then run
        license plate detection on the cropped frame. We can create BoundingBox for vehicle detection, then calculate
        relative BoundingBox for the license plate detection, which allows us to draw the license plate bounding box
        on the original frame.

        Args:
            bbox: BoundingBox to calculate relative coordinates to
        Returns:
            Relative BoundingBox
        """
        relative_bbox = BoundingBox((
            self.xmin + self.width * bbox.xmin,
            self.ymin + self.height * bbox.ymin,
            self.xmin + self.width * bbox.xmax,
            self.ymin + self.height * bbox.ymax,
        ))
        return relative_bbox

    def map_point(self, x: float, y: float) -> Point:
        """
        Useful when you have a point inside the bounding box, and you want to map it to the frame.
        Example: You run face detection, create BoundingBox from the result, and also run
        facial landmarks detection on the cropped frame of the face. The landmarks are relative
        to the face bounding box, but you want to draw them on the original frame.

        Args:
            x: x coordinate of the point inside the bounding box (0..1)
            y: y coordinate of the point inside the bounding box (0..1)
        Returns:
            Point in absolute coordinates (0..1)
        """
        mapped_x, mapped_y = self.xmin + self.width * x, self.ymin + self.height * y
        return Point(mapped_x, mapped_y)

    def get_centroid(self) -> Point:
        """
        Returns the centroid of the bounding box.
        """
        return Point((self.xmin + self.xmax) / 2, (self.ymin + self.ymax) / 2)

    def denormalize(self, frame_shape: Sequence) -> Tuple[Tuple[int, int], Tuple[int, int]]:
        """
        Denormalize the bounding box to pixel coordinates (0..frame width, 0..frame height).
        Useful when you want to draw the bounding box on the frame.

        Args:
            frame_shape: Shape of the frame (height, width)

        Returns:
            Tuple of two points (top-left, bottom-right) in pixel coordinates
        """
        return (
            (int(frame_shape[1] * self.xmin), int(frame_shape[0] * self.ymin)),
            (int(frame_shape[1] * self.xmax), int(frame_shape[0] * self.ymax))
        )

    def add_padding(self, padding_fraction, relative_box: Optional['BoundingBox'] = None) -> 'BoundingBox':
        """
        Adds padding to the bounding box and calculates new relative coordinates.

        Example: We run face detection on a frame, but our facial landmark detection model was trained on head images, not
        face images. So we need to pad an image a bit (from face bounding box) to get a head bounding box. As a note, we already
        support padding for MultiStageNN pipelines. So now that we have relative facial landmarks, we want to draw them to the
        original (high-res) frame. We create bounding box from the face detection result, then add padding to it (this function), and then
        we can map the relative landmarks (using map_point_to_pixels()) to the original frame.
        """
        box = relative_box if relative_box else self
        return BoundingBox((
            max(0, self.xmin - box.width * padding_fraction),
            max(0, self.ymin - box.height * padding_fraction),
            min(1, self.xmax + box.width * padding_fraction),
            min(1, self.ymax + box.height * padding_fraction)
        ))

    def crop_frame(self, frame: np.ndarray) -> np.ndarray:
        """
        Crops the frame to the bounding box coordinates.
        """
        (top, left), (bottom, right) = self.denormalize(frame.shape)
        return frame[left:right, top:bottom]

    def resize_to_aspect_ratio(self,
                               old_aspect_ratio: Union[float, Sequence],
                               new_aspect_ratio: Union[float, Sequence],
                               resize_mode: Union[ResizeMode, str] = ResizeMode.LETTERBOX) -> 'BoundingBox':
        """
        Calculates a new BoundingBox, based on the current BoundingBox, but with a different aspect ratio.
        Example: If the original frame is 1920x1080, and we have a model with input size of 300x300,
        the aspect ratio is 1:1, but the frame is 16:9. In that case, we can use one of four Resize modes,
        to adjust the frame to the new aspect ratio:
        - ResizeMode.LETTERBOX: Letterbox the frame (apply padding)
        - ResizeMode.STRETCH: Stretch the frame
        - ResizeMode.CROP: Crop a bit of the frame to match the aspect ratio, then scale it to the new size
        - ResizeMode.FULL_CROP: Middle-crop the frame to the required frame size (without scaling)
        This function will map the inference result (done on 300x300 frame) to the original frame.

        Args:
            old_aspect_ratio: Original aspect ratio (of the full frame)
            new_aspect_ratio: Aspect ratio of the frame we want to target
            resize_mode: Resize mode to use
        """
        if isinstance(resize_mode, str):
            resize_mode = ResizeMode.parse(resize_mode)
        if isinstance(old_aspect_ratio, (Sequence, np.ndarray)):
            # width / height. Here, we usually pass frame.shape, which is (height, width, channels)
            old_aspect_ratio = old_aspect_ratio[1] / old_aspect_ratio[0]
        if isinstance(new_aspect_ratio, (Sequence, np.ndarray)):
            # width / height
            new_aspect_ratio = new_aspect_ratio[0] / new_aspect_ratio[1]

        bb = BoundingBox((0, 0, 1, 1))

        if resize_mode == ResizeMode.LETTERBOX:
            if old_aspect_ratio < new_aspect_ratio:
                padding = (new_aspect_ratio - old_aspect_ratio) / (2 * old_aspect_ratio)
                bb = BoundingBox((-padding, 0, 1 + padding, 1))
            else:
                padding = (old_aspect_ratio - new_aspect_ratio) / (2 * new_aspect_ratio)
                bb = BoundingBox((0, -padding, 1, 1 + padding))
        elif resize_mode in [ResizeMode.CROP, ResizeMode.FULL_CROP]:
            if old_aspect_ratio < new_aspect_ratio:
                cropping = (1 - (old_aspect_ratio / new_aspect_ratio)) / 2
                bb = BoundingBox((0, cropping, 1, 1 - cropping))
            else:
                cropping = (1 - (new_aspect_ratio / old_aspect_ratio)) / 2
                bb = BoundingBox((cropping, 0, 1 - cropping, 1))
        else:
            # ResizeMode.STRETCH doesn't require any adjustments, as the aspect ratio is the same
            # So we will just keep the original "bb"
            pass

        return self.get_relative_bbox(bb)



import numpy as np
import depthai as dai
from depthai_sdk.classes.packets import SemanticSegmentation


def decode(nndata: dai.NNData):
    [print(f"Layer name: {l.name}, Type: {l.dataType}, Dimensions: {l.dims}") for l in data.getAllLayers()]
    # after squeeze the data.shape is 4,512, 896
    # output = np.squeeze(nndata.getLayerFp16('L0317_ReWeight_SoftMax'))
    # # Output is an array of 4x512x896 == 1835008 bytes, and has values from 0 to 1
    # output = output.reshape((4, 512, 896))
    # classes = np.asarray([0, 1, 2, 3], dtype=np.uint8)
    # indices = np.argmax(output, axis=0)
    # outputMap = np.take(classes, indices, axis=0)
    # return SemanticSegmentation(nndata, mask=outputMap)




import math
from typing import List, Tuple, Dict, Any, Optional, Union

import depthai as dai

monoResolutions: Dict[dai.MonoCameraProperties.SensorResolution, Tuple[int, int]] = {
    dai.MonoCameraProperties.SensorResolution.THE_1200_P: (1920, 1200),  # Monochrome AR0234
    dai.MonoCameraProperties.SensorResolution.THE_800_P: (1280, 800),  # OV9282
    dai.MonoCameraProperties.SensorResolution.THE_720_P: (1280, 720),
    dai.MonoCameraProperties.SensorResolution.THE_480_P: (640, 480),  # OV7251
    dai.MonoCameraProperties.SensorResolution.THE_400_P: (640, 400),
}

colorResolutions: Dict[dai.ColorCameraProperties.SensorResolution, Tuple[int, int]] = {
    dai.ColorCameraProperties.SensorResolution.THE_5312X6000: (5312, 6000),  # IMX582 cropped
    dai.ColorCameraProperties.SensorResolution.THE_13_MP: (4208, 3120),  # AR214
    dai.ColorCameraProperties.SensorResolution.THE_12_MP: (4056, 3040),  # IMX378, IMX477, IMX577
    dai.ColorCameraProperties.SensorResolution.THE_4000X3000: (4000, 3000),  # IMX582 with binning enabled
    dai.ColorCameraProperties.SensorResolution.THE_4_K: (3840, 2160),
    dai.ColorCameraProperties.SensorResolution.THE_1200_P: (1920, 1200),  # AR0234
    dai.ColorCameraProperties.SensorResolution.THE_1080_P: (1920, 1080),
    dai.ColorCameraProperties.SensorResolution.THE_1440X1080: (1440, 1080),
    dai.ColorCameraProperties.SensorResolution.THE_5_MP: (2592, 1944),  # OV5645
    dai.ColorCameraProperties.SensorResolution.THE_800_P: (1280, 800),  # OV9782
    dai.ColorCameraProperties.SensorResolution.THE_720_P: (1280, 720),
}

sensorResolutions: Dict[Any, Tuple[int, int]] = []
sensorResolutions.extend(monoResolutions)
sensorResolutions.extend(colorResolutions)


def availableIspScales() -> List[Tuple[int, Tuple[int, int]]]:
    """
    Calculates all supported
    @rtype: List[ratio, [Numerator, Denominator]]
    """
    lst = []
    for n in range(1, 16 + 1):
        for d in range(n, 63 + 1):
            # Chroma needs 2x extra downscaling
            if d < 32 or n % 2 == 0:
                # Only if irreducible
                if math.gcd(n, d) == 1:
                    lst.append((n / d, (n, d)))
    lst.sort(reverse=True)
    return lst


def getClosestVideoSize(width: int, height: int, videoEncoder: bool = False) -> Tuple[int, int]:
    """
    For colorCamera.video output
    """
    width_divider = 32 if videoEncoder else 2
    width = (width // width_divider) * width_divider

    height_divider = 8 if videoEncoder else 2
    height = (height // height_divider) * height_divider

    return (width, height)


def getClosestIspScale(camResolution: Tuple[int, int],
                       width: Optional[int] = None,
                       height: Optional[int] = None,
                       videoEncoder: bool = False,
                       encoderFlag: bool = True,
                       ) -> List[int]:
    """
    Provides the closest ISP scaling values to either specified width or height.
    @param camResolution: Resolution (W, H) of the ColorCamera
    @param width: Desired width after ISP scaling. Conflicts with height
    @param height: Desired height after ISP scaling. Conflicts with width
    @param videoEncoder: If we want to stream ISP output directly into VideoEncoder (take into account its limitations).
    Width and height scaling values won't be the same (ignore aspect ratio), as otherwise only 5 ISP scaling options are
    available.
    @param encoderFlag: Not for user. Flag to avoid infinite looping.
    @return ISP scaling values (list of 4 ints)
    """
    if width and height:
        raise ValueError("You have to specify EITHER width OR height to calculate desired ISP scaling options!")
    if width is None and height is None:
        raise ValueError("You have to provide width or height calculate desired ISP scaling options!")

    minError = 999999
    ispScale: List[int] = None
    for ratio, (n, d) in availableIspScales():
        newW = int((camResolution[0] * n - 1) / d + 1)
        newH = int((camResolution[1] * n - 1) / d + 1)

        if (videoEncoder and
                (width and newW % 32 != 0 or
                 height and newH % 8 != 0)):
            continue  # ISP output size isn't supported by VideoEncoder

        # Currently, new ISP width must be divisible by 2.
        # TODO: Width even number requirement was temporarily added because sometimes images were very zoomed. Engineers
        # are looking into it, and we will remove this check once this is fixed in firmware.
        if newW % 2 != 0 or newH % 2 != 0:
            continue

        err = abs((newW - width) if width is not None else (newH - height))
        if err < minError:
            ispScale = [n, d, n, d]
            minError = err

    if videoEncoder and encoderFlag:
        # Calculate the ISP scale for the other direction. Note that this means aspect ratio won't be preserved
        if width:
            hScale = getClosestIspScale(camResolution,
                                        height=int(camResolution[1] * ispScale[0] / ispScale[1]),
                                        videoEncoder=True,
                                        encoderFlag=False  # To avoid infinite loop
                                        )
            ispScale[2] = hScale[2]
            ispScale[3] = hScale[3]
        else:
            wScale = getClosestIspScale(camResolution,
                                        width=int(camResolution[0] * ispScale[0] / ispScale[1]),
                                        videoEncoder=True,
                                        encoderFlag=False  # To avoid infinite loop
                                        )
            ispScale[0] = wScale[0]
            ispScale[1] = wScale[1]

    return ispScale


def setCameraControl(control: dai.CameraControl,
                     manualFocus: Optional[int] = None,
                     afMode: Optional[dai.CameraControl.AutoFocusMode] = None,
                     awbMode: Optional[dai.CameraControl.AutoWhiteBalanceMode] = None,
                     sceneMode: Optional[dai.CameraControl.SceneMode] = None,
                     antiBandingMode: Optional[dai.CameraControl.AntiBandingMode] = None,
                     effectMode: Optional[dai.CameraControl.EffectMode] = None,
                     sharpness: Optional[int] = None,
                     lumaDenoise: Optional[int] = None,
                     chromaDenoise: Optional[int] = None,
                     ):
    """
    This function will be used when initializing cameras (ColorCamera.initialControl) and during runtime,
    when user configures camera with keyboard (if enabled).
    """
    if manualFocus is not None:
        control.setManualFocus(manualFocus)
    if afMode is not None:
        control.setAutoFocusMode(afMode)
    if awbMode is not None:
        control.setAutoWhiteBalanceMode(awbMode)
    if sceneMode is not None:
        control.setSceneMode(sceneMode)
    if antiBandingMode is not None:
        control.setAntiBandingMode(antiBandingMode)
    if effectMode is not None:
        control.setEffectMode(effectMode)
    if sharpness is not None:
        control.setSharpness(sharpness)
    if lumaDenoise is not None:
        control.setLumaDenoise(lumaDenoise)
    if chromaDenoise is not None:
        control.setChromaDenoise(chromaDenoise)

    # TODO: Add contrast, exposure compensation, brightness, manual exposure, and saturation


def get_max_resolution(node: dai.node, sensor: dai.CameraFeatures) -> Union[
    dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution]:
    max_res = None
    max_num = 0
    for conf in sensor.configs:
        if node == dai.node.ColorCamera and conf.type != dai.CameraSensorType.COLOR:
            continue
        if node == dai.node.MonoCamera and conf.type != dai.CameraSensorType.MONO:
            continue
        (res, size) = get_sensor_resolution(conf.type, conf.width, conf.height)
        if size[0] * size[1] > max_num:
            max_num = size[0] * size[1]
            max_res = res
    return max_res


def get_sensor_resolution(type: dai.CameraSensorType, width: int, height: int) -> Tuple[
    Union[dai.ColorCameraProperties.SensorResolution, dai.MonoCameraProperties.SensorResolution], Tuple[int, int]]:
    def get_res(resolutions: Dict[Any, Tuple[int, int]]):
        for res, (w, h) in resolutions.items():
            if width == w and height == h:
                return (res, (w, h))

    if type == dai.CameraSensorType.COLOR:
        return get_res(colorResolutions)
    elif type == dai.CameraSensorType.MONO:
        return get_res(monoResolutions)
    else:
        raise Exception('Camera sensor type unknown!', type)


def get_resolution_size(
        resolution: Union[
            dai.ColorCameraProperties.SensorResolution,
            dai.MonoCameraProperties.SensorResolution
        ]) -> Tuple[int, int]:
    if resolution in colorResolutions:
        return colorResolutions[resolution]
    elif resolution in monoResolutions:
        return monoResolutions[resolution]
    else:
        raise Exception('Camera sensor resolution unknown!', resolution)


def getClosesResolution(sensor: dai.CameraFeatures,
                        type: dai.CameraSensorType,
                        width: Optional[int] = None,
                        height: Optional[int] = None, ):
    if width and height:
        raise ValueError("You have to specify EITHER width OR height to calculate desired ISP scaling options!")
    if width is None and height is None:
        raise ValueError("You have to provide width or height calculate desired ISP scaling options!")

    minError = 999999
    closestRes = None
    desired, i = (width, 0) if width is not None else (height, 1)

    resolutions = [get_sensor_resolution(type, conf.width, conf.height) for conf in sensor.configs if conf.type == type]
    resolutions = [res for res in resolutions if res is not None]

    for (res, size) in resolutions:
        err = abs(size[i] - desired)
        if err < minError:
            minError = err
            closestRes = res
    return closestRes


def getResize(size: Tuple[int, int],
              width: Optional[int] = None,
              height: Optional[int] = None) -> Tuple[int, int]:
    if width and height:
        raise ValueError("You have to specify EITHER width OR height to calculate desired ISP scaling options!")
    if width is None and height is None:
        raise ValueError("You have to provide width or height calculate desired ISP scaling options!")

    if width:
        return width, int(size[1] / size[0] * width)
    else:
        return int(size[0] / size[1] * height), height



from abc import ABC, abstractmethod
from datetime import timedelta
from typing import Sequence, Tuple, List, Union, Optional, Dict, Callable

import depthai as dai
import numpy as np

from depthai_sdk.classes import ImgLandmarks, SemanticSegmentation
from depthai_sdk.classes.nn_results import Detection, TrackingDetection, TwoStageDetection
from depthai_sdk.visualize.bbox import BoundingBox
from depthai_sdk.visualize.configs import StereoColor, TextPosition
from depthai_sdk.visualize.visualizer import Visualizer

try:
    import cv2
except ImportError:
    cv2 = None


class BasePacket(ABC):
    """
    Base class for all packets.
    """

    def __init__(self, name: str):
        self.name = name

    def prepare_visualizer_objects(self, visualizer: 'Visualizer') -> None:
        """
        Prepare visualizer objects (boxes, lines, text, etc.), so visualizer can draw them on the frame.

        Args:
            visualizer: Visualizer object.
        """
        pass

    @abstractmethod
    def get_timestamp(self) -> timedelta:
        raise NotImplementedError()

    @abstractmethod
    def get_sequence_num(self) -> int:
        raise NotImplementedError()


class NNDataPacket(BasePacket):
    """
    Contains only dai.NNData message
    """

    def __init__(self, name: str, nn_data: dai.NNData):
        self.msg = nn_data
        super().__init__(name)

    def get_timestamp(self) -> timedelta:
        return self.msg.getTimestamp()

    def get_sequence_num(self) -> int:
        return self.msg.getTimestampDevice()


class FramePacket(BasePacket):
    """
    Contains only dai.ImgFrame message and cv2 frame, which is used by visualization logic.
    """

    def __init__(self,
                 name: str,
                 msg: dai.ImgFrame,
                 ):
        self.msg = msg
        self._get_codec = None
        self.__frame = None
        super().__init__(name)

    @property
    def frame(self):
        if self.__frame is None:
            self.__frame = self.decode()
        return self.__frame

    def get_timestamp(self) -> timedelta:
        return self.msg.getTimestampDevice(dai.CameraExposureOffset.MIDDLE)

    def get_sequence_num(self) -> int:
        return self.msg.getSequenceNum()

    def set_decode_codec(self, get_codec: Callable):
        self._get_codec = get_codec

    def decode(self) -> Optional[np.ndarray]:
        if self._get_codec is None:
            return self.msg.getCvFrame() if cv2 else None

        codec = self._get_codec()
        if codec is None:
            return self.msg.getCvFrame() if cv2 else None

        # PyAV decoding support H264, H265, JPEG and Lossless JPEG
        enc_packets = codec.parse(self.msg.getData())
        if len(enc_packets) == 0:
            return None

        frames = codec.decode(enc_packets[-1])
        if not frames:
            return None

        return frames[0].to_ndarray(format='bgr24')

    def get_size(self) -> Tuple[int, int]:
        return self.msg.getWidth(), self.msg.getHeight()


class DisparityPacket(FramePacket):
    def __init__(self,
                 name: str,
                 img: dai.ImgFrame,
                 multiplier: float,
                 disparity_map: Optional[np.ndarray] = None,
                 colorize: StereoColor = None,
                 colormap: int = None,
                 aligned_frame: Optional[dai.ImgFrame] = None,
                 confidence_map: Optional[np.ndarray] = None
                 ):
        """
        disparity_map might be filtered, eg. if WLS filter is enabled
        """
        super().__init__(name=name, msg=img)
        self.aligned_frame = aligned_frame
        self.disparity_map = disparity_map
        self.multiplier = multiplier
        self.colorize = colorize
        self.colormap = colormap

        self.confidence_map = confidence_map
        self.depth_score = None
        if self.confidence_map:
            values = 1 - (self.confidence_map.getData() / 255)
            values_no_outliers = values[np.logical_and(values > 0.0, values < 1.0)]
            self.depth_score = np.mean(values_no_outliers)

    def get_disparity(self) -> np.ndarray:
        if self.disparity_map is not None:
            return self.disparity_map
        else:
            self.msg.getFrame()

    def get_colorized_frame(self, visualizer) -> np.ndarray:
        frame = self.get_disparity()
        colorized_disp = frame * self.multiplier

        try:
            aligned_frame = self.aligned_frame.getCvFrame()
        except AttributeError:
            aligned_frame = None

        stereo_config = visualizer.config.stereo

        colorize = self.colorize or stereo_config.colorize
        if self.colormap is not None:
            colormap = self.colormap
        else:
            colormap = stereo_config.colormap
            colormap[0] = [0, 0, 0]  # Invalidate pixels 0 to be black

        if aligned_frame is not None and colorized_disp.ndim == 2 and aligned_frame.ndim == 3:
            colorized_disp = colorized_disp[..., np.newaxis]

        if colorize == StereoColor.GRAY:
            pass
        elif colorize == StereoColor.RGB:
            colorized_disp = cv2.applyColorMap(colorized_disp.astype(np.uint8), colormap)
        elif colorize == StereoColor.RGBD:
            colorized_disp = cv2.applyColorMap(
                (colorized_disp + aligned_frame * 0.5).astype(np.uint8), colormap
            )
        return colorized_disp


class DepthPacket(FramePacket):
    def __init__(self, name: str,
                 msg: dai.ImgFrame):
        super().__init__(name, msg)
        self.depth = msg.getFrame()


class DisparityDepthPacket(DisparityPacket):
    def __init__(self,
                 name: str,
                 img_frame: dai.ImgFrame,
                 colorize: StereoColor = None,
                 colormap: int = None,
                 aligned_frame: Optional[dai.ImgFrame] = None,
                 disp_scale_factor=255 / 95,
                 confidence_map=None
                 ):
        # DepthPacket.__init__(self, name=name, msg=img_frame)
        super().__init__(
            name=name,
            img=img_frame,
            disparity_map=None,
            multiplier=255 / 95,
            colorize=colorize,
            colormap=colormap,
            aligned_frame=aligned_frame,
            confidence_map=confidence_map
        )
        self.disp_scale_factor = disp_scale_factor

    def get_disparity(self) -> np.ndarray:
        with np.errstate(divide='ignore'):
            disparity = self.disp_scale_factor / self.msg.getFrame()
        disparity[disparity == np.inf] = 0
        return disparity

    # def get_colorized_frame(self, visualizer) -> np.ndarray:
    # Convert depth to disparity for nicer visualization


class PointcloudPacket(BasePacket):
    def __init__(self,
                 name: str,
                 points: np.ndarray,
                 depth_map: dai.ImgFrame,
                 colorize_frame: Optional[dai.ImgFrame]):
        super().__init__(name=name)
        self.points = points
        self.colorize_frame = colorize_frame.getCvFrame() if colorize_frame is not None else None
        self.depth_map = depth_map

    def get_sequence_num(self) -> int:
        return self.depth_map.getSequenceNum()

    def get_timestamp(self) -> timedelta:
        return self.depth_map.getTimestampDevice()

    def crop_points(self, bb: BoundingBox) -> np.ndarray:
        """
        Crop points to the bounding box

        Returns: Cropped section of the point cloud
        """
        x1, y1, x2, y2 = bb.to_tuple(self.points.shape)
        return self.points[y1:y2, x1:x2]


class SpatialBbMappingPacket(DisparityDepthPacket):
    """
    Output from Spatial Detection nodes - depth frame + bounding box mappings. Inherits FramePacket.
    """

    def __init__(self,
                 name: str,
                 msg: dai.ImgFrame,
                 spatials: dai.SpatialImgDetections,
                 disp_scale_factor: float):
        super().__init__(name=name,
                         img_frame=msg,
                         disp_scale_factor=disp_scale_factor)
        self.spatials = spatials

    def prepare_visualizer_objects(self, vis: Visualizer) -> None:
        # Add detections to packet
        for detection in self.spatials.detections:
            br = detection.boundingBoxMapping.roi.bottomRight()
            tl = detection.boundingBoxMapping.roi.topLeft()
            bbox = BoundingBox([tl.x, tl.y, br.x, br.y])
            # Add detections to visualizer
            vis.add_bbox(
                bbox=bbox,
                thickness=3,
                color=(0, 0, 0)
            )
            vis.add_bbox(
                bbox=bbox,
                thickness=1,
                color=(255, 255, 255)
            )


class NnOutputPacket(FramePacket):
    """
    NN result + image frame. Inherits FramePacket.
    """

    def __init__(self,
                 name: str,
                 msg: dai.ImgFrame,
                 nn_data: dai.NNData,
                 bbox: BoundingBox
                 ):
        super().__init__(name=name,
                         msg=msg)
        self.nn_data = nn_data
        self.bbox = bbox


class ImgLandmarksPacket(NnOutputPacket):
    """
    Output from Landmarks Estimation nodes - image frame + landmarks. Inherits NnOutputPacket.
    """

    def __init__(self,
                 name: str,
                 msg: dai.ImgFrame,
                 nn_data: dai.NNData,
                 landmarks: ImgLandmarks,
                 bbox: BoundingBox):
        super().__init__(name=name,
                         msg=msg,
                         nn_data=nn_data,
                         bbox=bbox)
        self.landmarks = landmarks

    def prepare_visualizer_objects(self, vis: Visualizer) -> None:
        all_landmarks = self.landmarks.landmarks
        all_landmarks_indices = self.landmarks.landmarks_indices
        colors = self.landmarks.colors
        w, h = self.get_size()
        for landmarks, indices in zip(all_landmarks, all_landmarks_indices):
            for i, landmark in enumerate(landmarks):
                # Map normalized coordinates to frame coordinates
                l = [(int(point[0] * w), int(point[1] * h)) for point in landmark]
                idx = indices[i]

                vis.add_line(pt1=tuple(l[0]), pt2=tuple(l[1]), color=colors[idx], thickness=4)
                vis.add_circle(coords=tuple(l[0]), radius=8, color=colors[idx], thickness=-1)
                vis.add_circle(coords=tuple(l[1]), radius=8, color=colors[idx], thickness=-1)


class SemanticSegmentationPacket(NnOutputPacket):
    """
    Output from Semantic Segmentation nodes - image frame + segmentation mask. Inherits NnOutputPacket.
    """

    def __init__(self,
                 name: str,
                 msg: dai.ImgFrame,
                 nn_data: dai.NNData,
                 segmentation: SemanticSegmentation,
                 bbox: BoundingBox):
        super().__init__(name=name,
                         msg=msg,
                         nn_data=nn_data,
                         bbox=bbox)
        self.segmentation = segmentation

    def prepare_visualizer_objects(self, vis: Visualizer) -> None:
        raise NotImplementedError('Semantic segmentation visualization is not implemented yet!')


class DetectionPacket(FramePacket):
    """
    Output from Detection Network nodes - image frame + image detections. Inherits FramePacket.
    """

    def __init__(self,
                 name: str,
                 msg: dai.ImgFrame,
                 dai_msg: Union[dai.ImgDetections, dai.SpatialImgDetections, dai.NNData],
                 bbox: BoundingBox,
                 ):

        super().__init__(name=name,
                         msg=msg)

        self.img_detections = dai_msg
        self.bbox = bbox
        self.detections: List[Detection] = []

    def _is_spatial_detection(self) -> bool:
        return isinstance(self.img_detections, dai.SpatialImgDetections)

    def prepare_visualizer_objects(self, vis: Visualizer) -> None:
        # Add detections to packet
        for detection in self.detections:
            # Add detections to visualizer
            vis.add_bbox(
                bbox=self.bbox.get_relative_bbox(detection.bbox),
                # label=detection.label_str,
                color=detection.color,
            )
            vis.add_text(
                f'{detection.label_str} {100 * detection.confidence:.0f}%',
                bbox=self.bbox.get_relative_bbox(detection.bbox),
                position=TextPosition.TOP_LEFT,
            )
            # Spatial coordinates
            if type(detection.img_detection) == dai.SpatialImgDetection:
                x_meters = detection.img_detection.spatialCoordinates.x / 1000
                y_meters = detection.img_detection.spatialCoordinates.y / 1000
                z_meters = detection.img_detection.spatialCoordinates.z / 1000
                vis.add_text(
                    f'X: {x_meters:.2f}m\nY: {y_meters:.2f}m\nZ: {z_meters:.2f}m',
                    bbox=self.bbox.get_relative_bbox(detection.bbox),
                    position=TextPosition.BOTTOM_LEFT,
                )


class TrackerPacket(FramePacket):
    """
    Output of Object Tracker node. Tracklets + Image frame. Inherits FramePacket.
    """

    def __init__(self,
                 name: str,
                 msg: dai.ImgFrame,
                 tracklets: dai.Tracklets,
                 bbox: BoundingBox,
                 ):
        super().__init__(name=name,
                         msg=msg)

        # int: object_id, list: TrackingDetection
        self.tracklets: Dict[int, List[TrackingDetection]] = {}
        self.daiTracklets = tracklets
        self.bbox = bbox

    def _is_spatial_detection(self) -> bool:
        coords = self.daiTracklets.tracklets[0].spatialCoordinates
        return coords.x != 0.0 or coords.y != 0.0 or coords.z != 0.0

    def prepare_visualizer_objects(self, visualizer: Visualizer) -> None:
        tracking_config = visualizer.config.tracking
        for obj_id, tracking_dets in self.tracklets.items():
            tracking_det = tracking_dets[-1]  # Get the last detection
            bb = tracking_det.filtered_2d or tracking_det.bbox
            visualizer.add_bbox(
                bbox=self.bbox.get_relative_bbox(bb),
                label=f"[{obj_id}] {tracking_det.label_str}",
                color=tracking_det.color,
            )
            visualizer.add_text(
                f'{tracking_det.label_str} {100 * tracking_det.confidence:.0f}%',
                bbox=self.bbox.get_relative_bbox(bb),
                position=TextPosition.TOP_LEFT,
            )
            if visualizer.config.tracking.show_speed and \
                    tracking_det.speed is not None:
                visualizer.add_text(
                    text=f"{tracking_det.speed:.2f} m/s",
                    color=tracking_det.color,
                    bbox=self.bbox.get_relative_bbox(bb),
                    position=TextPosition.BOTTOM_RIGHT,
                )
            w, h = self.get_size()
            tracklet_length = 0
            for i in reversed(range(len(tracking_dets) - 1)):
                p1 = self.bbox.get_relative_bbox(tracking_dets[i].bbox).get_centroid().denormalize((h, w))
                p2 = self.bbox.get_relative_bbox(tracking_dets[i + 1].bbox).get_centroid().denormalize((h, w))

                if tracking_config.max_length != -1:
                    tracklet_length += np.linalg.norm(np.array(p1) - np.array(p2))
                    if tracking_config.max_length < tracklet_length:
                        break

                thickness = tracking_config.line_thickness
                if tracking_config.fading_tails:
                    thickness = max(1, int(np.ceil(thickness * i / len(tracking_dets))))

                visualizer.add_line(pt1=p1, pt2=p2,
                                    color=tracking_dets[i].color,
                                    thickness=thickness
                                    )


class TwoStagePacket(DetectionPacket):
    """
    Output of 2-stage NN pipeline; Image frame, Image detections and multiple NNData results. Inherits DetectionPacket.
    """

    def __init__(self, name: str,
                 msg: dai.ImgFrame,
                 img_detections: dai.ImgDetections,
                 nn_data: List[dai.NNData],
                 labels: List[int],
                 bbox: BoundingBox
                 ):
        super().__init__(name=name,
                         msg=msg,
                         dai_msg=img_detections,
                         bbox=bbox
                         )
        self.nnData = nn_data
        self.labels = labels
        self._cntr = 0

    def _add_detection(self, img_det: dai.ImgDetection, bbox: np.ndarray, txt: str, color):
        det = TwoStageDetection()
        det.img_detection = img_det
        det.color = color
        det.top_left = (bbox[0], bbox[1])
        det.bottom_right = (bbox[2], bbox[3])

        # Append the second stage NN result to the detection
        if self.labels is None or img_det.label in self.labels:
            det.nn_data = self.nnData[self._cntr]
            self._cntr += 1

        self.detections.append(det)


class IMUPacket(BasePacket):
    def __init__(self, name, packet: dai.IMUPacket, rotation=None):
        self.packet = packet
        super().__init__(name)

        self.acceleroMeter = packet.acceleroMeter
        self.gyroscope = packet.gyroscope
        self.magneticField = packet.magneticField
        self.rotationVector = rotation if rotation is not None else packet.rotationVector

        # Check which reports are available
        self.available_reports: Dict[str, dai.IMUReport] = {}
        for i, val in enumerate([self.acceleroMeter, self.gyroscope, self.magneticField, self.rotationVector]):
            if (i == 3 and rotation) or val.getTimestampDevice() != timedelta(0):
                self.available_reports[val.__class__.__name__] = val

    def get_imu_vals(self) -> Tuple[Sequence, Sequence, Sequence, Sequence]:
        """
        Returns imu values in a tuple. Returns in format (accelerometer_values, gyroscope_values, quaternion, magnetometer_values)
        """
        return (
            [self.acceleroMeter.x, self.acceleroMeter.y, self.acceleroMeter.z],
            [self.gyroscope.x, self.gyroscope.y, self.gyroscope.z],
            [self.rotationVector.i, self.rotationVector.j, self.rotationVector.k, self.rotationVector.real],
            [self.magneticField.x, self.magneticField.y, self.magneticField.z]
        )

    def __str__(self):
        accelerometer_str = 'Accelerometer [m/s^2]: (x: %.2f, y: %.2f, z: %.2f)' % (
            self.packet.acceleroMeter.x,
            self.packet.acceleroMeter.y,
            self.packet.acceleroMeter.z
        )

        gyroscope_str = 'Gyroscope [rad/s]: (x: %.2f, y: %.2f, z: %.2f)' % (
            self.packet.gyroscope.x,
            self.packet.gyroscope.y,
            self.packet.gyroscope.z
        )

        return f'IMU Packet: {accelerometer_str} {gyroscope_str}'

    def _get_imu_report(self) -> dai.IMUReport:
        """
        Get the first available IMU report
        """
        for name, val in self.available_reports.items():
            return val

    def get_timestamp(self) -> timedelta:
        return self._get_imu_report().getTimestampDevice()

    def get_sequence_num(self) -> int:
        return self._get_imu_report().getSequenceNum()



from pathlib import Path
from typing import Dict

import depthai as dai
from mcap_ros1.writer import Writer as Ros1Writer

from depthai_sdk.logger import LOGGER
from depthai_sdk.integrations.ros.ros_base import RosBase
# from depthai_sdk.integrations.ros.depthai2ros import DepthAi2Ros1
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames
from depthai_sdk.recorders.abstract_recorder import *


class McapRecorder(Recorder, RosBase):
    """
    This is a helper class that lets you save frames into mcap (.mcap), which can be replayed using Foxglove studio app.
    """

    def __init__(self):
        self.path = None
        self.converter = None
        self.stream = None
        self.ros_writer = None

        self._closed = False
        self._pcl = False
        self._stream_type: Dict[str, OakStream] = dict()
        self._name_mapping: Dict[str, str] = dict()  # XLink name to nice name mapping

    def update(self, path: Path, device: dai.Device, xouts: List['XoutFrames']):
        """
        Args:
            path (Path): Path to which we record
            device (dai.Device): OAK Device
            xouts (List['XoutFrames']): List of outputs, which are used to record
        """
        self.path = str(path / "recordings.mcap")
        self.converter = Bridge(device)
        self.stream = open(self.path, "w+b")
        self.ros_writer = Ros1Writer(output=self.stream)

        RosBase.__init__(self)
        RosBase.update(self, device, xouts)

        for xout in xouts:
            if xout.is_h26x():
                raise Exception("MCAP recording only supports MJPEG encoding!")
            if xout.is_mjpeg() and xout.lossless:
                # Foxglove Studio doesn't (yet?) support Lossless MJPEG
                raise Exception("MCAP recording doesn't support Lossless MJPEG encoding!")

    def write(self, name: str, frame: dai.ImgFrame):
        self.new_msg(name, frame)  # To RosMsg which arrives to new_ros_msg()

    def new_ros_msg(self, topic: str, ros_msg):
        self.ros_writer.write_message(topic, ros_msg)

    def set_pointcloud(self, enable: bool):
        """
        Whether to convert depth to pointcloud
        """
        self._pcl = enable

    # def write(self, name: str, frame: dai.ImgFrame):
    #     if name not in self._name_mapping:
    #         return
    #
    #     name = self._name_mapping[name]
    #     if self._stream_type[name].is_depth() and self._pcl:
    #         msg = self.converter.PointCloud2(frame)
    #         self.ros_writer.write_message(f"pointcloud/raw", msg)
    #         # tf = self.converter.TfMessage(frame)
    #         # self.ros_writer.write_message(f"pointcloud/tf", tf)
    #     elif self._stream_type[name].is_mjpeg():
    #         msg = self.converter.CompressedImage(frame)
    #         self.ros_writer.write_message(f"{name}/compressed", msg)
    #     elif self._stream_type[name].is_imu():
    #         frame: dai.IMUData
    #         for imu_packet in frame.packets:
    #             msg = self.converter.Imu(imu_packet)
    #             self.ros_writer.write_message(f"imu", msg)
    #     else:  # Non-encoded frame; rgb/mono/depth
    #         msg = self.converter.Image(frame)
    #         self.ros_writer.write_message(f"{name}/raw", msg)

    def close(self) -> None:
        if self._closed: return
        self._closed = True
        self.ros_writer.finish()
        self.stream.close()
        LOGGER.info(f'.MCAP recording saved at {self.path}')



from abc import abstractmethod
from typing import List, Union, Dict

from depthai_sdk.oak_outputs.syncing import SequenceNumSync
from depthai_sdk.oak_outputs.xout.xout_base import XoutBase, StreamXout


class XoutSeqSync(XoutBase, SequenceNumSync):
    def xstreams(self) -> List[StreamXout]:
        return self.streams

    def __init__(self, streams: List[StreamXout]):
        # Filter out None streams
        self.streams = [s for s in streams if s is not None]

        # Save StreamXout before initializing super()!
        XoutBase.__init__(self)
        SequenceNumSync.__init__(self, len(self.streams))
        self.msgs = dict()

    @abstractmethod
    def package(self, msgs: Union[List, Dict]):
        raise NotImplementedError('XoutSeqSync is an abstract class, you need to override package() method!')

    def new_msg(self, name: str, msg):
        # Ignore frames that we aren't listening for
        if name not in self._streams: return

        synced = self.sync(msg.getSequenceNum(), name, msg)
        if synced:
            return self.package(synced)



import time
from abc import ABC
from collections import deque, defaultdict
from pathlib import Path
from typing import Dict


class BaseWriter(ABC):
    def __init__(self, path: Path, name: str):
        self.path = path
        self.name = name

        self._buffers: Dict[str, deque] = {}
        self._buffers_max_seconds: Dict[str, int] = {}  # in seconds
        self._buffers_timestamps = defaultdict(list)
        self._buffers_approx_fps: Dict[str, float] = {}
        self._file = None

    def create_file_for_buffer(self, subfolder: str, bufname: str):
        raise NotImplementedError()

    def init_buffer(self, name: str, max_seconds: int):
        if max_seconds > 0:
            self._buffers[name] = deque()
            self._buffers_max_seconds[name] = max_seconds

    def add_to_buffer(self, name: str, frame):
        if self._buffers[name] is None:
            return

        timestamp = time.time()
        self._buffers_timestamps[name].append(timestamp)

        # Calculate time window based on max_seconds
        time_window = self._buffers_max_seconds[name]

        # Remove frames that fall outside the time window
        while self._buffers_timestamps[name] and (timestamp - self._buffers_timestamps[name][0] > time_window):
            self._buffers[name].popleft()
            self._buffers_timestamps[name].pop(0)

        self._buffers[name].append(frame)

    def is_buffer_full(self, name: str) -> bool:
        if self._buffers[name].maxlen:
            return len(self._buffers[name]) == self._buffers[name].maxlen

        if not self._buffers_timestamps[name]:
            return False

        diff = self._buffers_timestamps[name][0] + self._buffers_max_seconds[name] - self._buffers_timestamps[name][-1]
        return diff < 0.1

    def is_buffer_empty(self, name: str) -> bool:
        return len(self._buffers[name]) == 0

    def write_from_buffer(self, name: str, n_elems: int):
        while len(self._buffers[name]) > 0 and n_elems > 0:
            frame = self._buffers[name].popleft()
            self.write(frame)
            n_elems -= 1

    def write(self, frame):
        raise NotImplementedError()

    def close(self):
        raise NotImplementedError()



from enum import IntEnum
from typing import Tuple, Union, List, Any, Dict

from depthai_sdk.classes.nn_results import TrackingDetection, TwoStageDetection
from depthai_sdk.visualize.configs import BboxStyle
from depthai_sdk.visualize.objects import VisBoundingBox

try:
    import cv2
except ImportError:
    cv2 = None

import depthai as dai
import numpy as np

from depthai_sdk.classes.packets import (
    DetectionPacket,
    SpatialBbMappingPacket,
    TrackerPacket,
)
from depthai_sdk.visualize.bbox import BoundingBox


class FramePosition(IntEnum):
    """
    Where on frame do we want to print text.
    """
    TopLeft = 0
    MidLeft = 1
    BottomLeft = 2
    TopMid = 10
    Mid = 11
    BottomMid = 12
    TopRight = 20
    MidRight = 21
    BottomRight = 22


if cv2 is not None:
    default_color_map = cv2.applyColorMap(np.arange(256, dtype=np.uint8), cv2.COLORMAP_TURBO)
    default_color_map[0] = [0, 0, 0]
else:
    default_color_map = None


class VisualizerHelper:
    bg_color = (0, 0, 0)
    front_color = (255, 255, 255)
    text_type = 0  # cv2.FONT_HERSHEY_SIMPLEX
    line_type = 16  # cv2.LINE_AA

    @classmethod
    def putText(cls,
                frame: np.ndarray,
                text: str,
                coords: Tuple[int, int],
                scale: float = 1.0,
                backColor: Tuple[int, int, int] = None,
                color: Tuple[int, int, int] = None):
        # Background text

        cv2.putText(frame, text, coords, cls.text_type, scale,
                    color=backColor if backColor else cls.bg_color,
                    thickness=int(scale * 3),
                    lineType=cls.line_type)
        # Front text
        cv2.putText(frame, text, coords, cls.text_type, scale,
                    color=(int(color[0]), int(color[1]), int(color[2])) if color else cls.front_color,
                    thickness=int(scale),
                    lineType=cls.line_type)

    @classmethod
    def line(cls,
             frame: np.ndarray,
             p1: Tuple[int, int], p2: Tuple[int, int],
             color=None,
             thickness: int = 1) -> None:
        cv2.line(frame, p1, p2,
                 cls.bg_color,
                 thickness * 3,
                 cls.line_type)
        cv2.line(frame, p1, p2,
                 (int(color[0]), int(color[1]), int(color[2])) if color else cls.front_color,
                 thickness,
                 cls.line_type)

    @classmethod
    def print_on_roi(cls, frame, topLeft, bottomRight, text: str, position: FramePosition = FramePosition.BottomLeft,
                     padPx=10):
        frame_roi = frame[topLeft[1]:bottomRight[1], topLeft[0]:bottomRight[0]]
        cls.print(frame=frame_roi, text=text, position=position, padPx=padPx)

    @classmethod
    def print(cls, frame, text: str, position: FramePosition = FramePosition.BottomLeft, padPx=10):
        """
        Prints text on the frame.
        @param frame: Frame
        @param text: Text to be printed
        @param position: Where on frame we want to print the text
        @param padPx: Padding (in pixels)
        """
        text_size = cv2.getTextSize(text, VisualizerHelper.text_type, fontScale=1.0, thickness=1)[0]
        frame_w = frame.shape[1]
        frame_h = frame.shape[0]

        y_pos = int(position) % 10
        if y_pos == 0:  # Y Top
            y = text_size[1] + padPx
        elif y_pos == 1:  # Y Mid
            y = int(frame_h / 2) + int(text_size[1] / 2)
        else:  # y_pos == 2. Y Bottom
            y = frame_h - padPx

        x_pos = int(position) // 10
        if x_pos == 0:  # X Left
            x = padPx
        elif x_pos == 1:  # X Mid
            x = int(frame_w / 2) - int(text_size[0] / 2)
        else:  # x_pos == 2  # X Right
            x = frame_w - text_size[0] - padPx
        cls.putText(frame, text, (x, y))


# def rectangle(frame, bbox, color: Tuple[int, int, int] = None):
#     x1, y1, x2, y2 = bbox
#     cv2.rectangle(frame, (x1, y1), (x2, y2), bg_color, 3)
#     cv2.rectangle(frame, pt1=(x1, y1), pt2=(x2, y2),
#                   color=(int(color[0]), int(color[1]), int(color[2])) if color else front_color, thickness=1)


def rectangle(src,
              bbox: np.ndarray,
              color: Tuple[int, int, int],
              thickness=-1,
              radius=0.1,
              line_type=16,  # cv2.LINE_AA
              alpha=0.15):
    """
    Draw the rectangle (bounding box) on the frame.
    @param src: Frame
    @param top_left: Top left corner of the bounding box
    @param bottom_right: Bottom right corner of the bounding box
    @param color: Color of the rectangle
    @param thickness: Thickness of the rectangle. If -1, it will colorize the rectangle as well (with alpha)
    @param radius: Radius of the corners (for rounded rectangle)
    @param line_type: Line type for the rectangle
    @param alpha: Alpha for colorizing of the rectangle
    @return: Frame
    """
    top_left = (bbox[0], bbox[1])
    bottom_right = (bbox[2], bbox[3])
    top_right = (bottom_right[0], top_left[1])
    bottom_left = (top_left[0], bottom_right[1])

    height = abs(bottom_right[1] - top_left[1])

    if radius > 1:
        radius = 1

    corner_radius = int(radius * (height / 2))

    if thickness < 0:
        overlay = src.copy()

        # big rect
        top_left_main_rect = (int(top_left[0] + corner_radius), int(top_left[1]))
        bottom_right_main_rect = (int(bottom_right[0] - corner_radius), int(bottom_right[1]))

        top_left_rect_left = (top_left[0], top_left[1] + corner_radius)
        bottom_right_rect_left = (bottom_left[0] + corner_radius, bottom_left[1] - corner_radius)

        top_left_rect_right = (top_right[0] - corner_radius, top_right[1] + corner_radius)
        bottom_right_rect_right = (bottom_right[0], bottom_right[1] - corner_radius)

        all_rects = [
            [top_left_main_rect, bottom_right_main_rect],
            [top_left_rect_left, bottom_right_rect_left],
            [top_left_rect_right, bottom_right_rect_right]
        ]

        [cv2.rectangle(overlay, pt1=rect[0], pt2=rect[1], color=color, thickness=thickness) for rect in all_rects]

        cv2.ellipse(overlay, (top_left[0] + corner_radius, top_left[1] + corner_radius), (corner_radius, corner_radius),
                    180.0, 0,
                    90, color, thickness, line_type)
        cv2.ellipse(overlay, (top_right[0] - corner_radius, top_right[1] + corner_radius),
                    (corner_radius, corner_radius),
                    270.0, 0,
                    90, color, thickness, line_type)
        cv2.ellipse(overlay, (bottom_right[0] - corner_radius, bottom_right[1] - corner_radius),
                    (corner_radius, corner_radius), 0.0, 0, 90,
                    color, thickness, line_type)
        cv2.ellipse(overlay, (bottom_left[0] + corner_radius, bottom_left[1] - corner_radius),
                    (corner_radius, corner_radius), 90.0, 0,
                    90, color, thickness, line_type)

        cv2.ellipse(src, (top_left[0] + corner_radius, top_left[1] + corner_radius), (corner_radius, corner_radius),
                    180.0, 0, 90,
                    color, 2, line_type)
        cv2.ellipse(src, (top_right[0] - corner_radius, top_right[1] + corner_radius), (corner_radius, corner_radius),
                    270.0, 0, 90,
                    color, 2, line_type)
        cv2.ellipse(src, (bottom_right[0] - corner_radius, bottom_right[1] - corner_radius),
                    (corner_radius, corner_radius), 0.0, 0, 90,
                    color, 2, line_type)
        cv2.ellipse(src, (bottom_left[0] + corner_radius, bottom_left[1] - corner_radius),
                    (corner_radius, corner_radius),
                    90.0, 0, 90,
                    color, 2, line_type)

        cv2.addWeighted(overlay, alpha, src, 1 - alpha, 0, src)
    else:  # Don't fill the rectangle
        # draw straight lines
        cv2.line(src, (top_left[0] + corner_radius, top_left[1]), (top_right[0] - corner_radius, top_right[1]), color,
                 abs(thickness), line_type)
        cv2.line(src, (top_right[0], top_right[1] + corner_radius), (bottom_right[0], bottom_right[1] - corner_radius),
                 color, abs(thickness), line_type)
        cv2.line(src, (bottom_right[0] - corner_radius, bottom_left[1]),
                 (bottom_left[0] + corner_radius, bottom_right[1]),
                 color, abs(thickness), line_type)
        cv2.line(src, (bottom_left[0], bottom_left[1] - corner_radius), (top_left[0], top_left[1] + corner_radius),
                 color,
                 abs(thickness), line_type)

        # draw arcs
        cv2.ellipse(src, (top_left[0] + corner_radius, top_left[1] + corner_radius), (corner_radius, corner_radius),
                    180.0, 0, 90,
                    color, thickness, line_type)
        cv2.ellipse(src, (top_right[0] - corner_radius, top_right[1] + corner_radius), (corner_radius, corner_radius),
                    270.0, 0, 90,
                    color, thickness, line_type)
        cv2.ellipse(src, (bottom_right[0] - corner_radius, bottom_right[1] - corner_radius),
                    (corner_radius, corner_radius), 0.0, 0, 90,
                    color, thickness, line_type)
        cv2.ellipse(src, (bottom_left[0] + corner_radius, bottom_left[1] - corner_radius),
                    (corner_radius, corner_radius),
                    90.0, 0, 90,
                    color, thickness, line_type)

    return src


def draw_mappings(packet: SpatialBbMappingPacket):
    dets = packet.spatials.detections
    for det in dets:
        roi = det.boundingBoxMapping.roi
        roi = roi.denormalize(packet.frame.shape[1], packet.frame.shape[0])
        top_left = roi.topLeft()
        bottom_right = roi.bottomRight()
        x_min = int(top_left.x)
        y_min = int(top_left.y)
        x_max = int(bottom_right.x)
        y_max = int(bottom_right.y)

        cv2.rectangle(packet.frame, (x_min, y_min), (x_max, y_max), VisualizerHelper.bg_color, 3)
        cv2.rectangle(packet.frame, (x_min, y_min), (x_max, y_max), VisualizerHelper.front_color, 1)


def draw_detections(packet: Union[DetectionPacket, TwoStageDetection, TrackerPacket],
                    norm: BoundingBox,
                    label_map: List[Tuple[str, Tuple]] = None):
    """
    Draw object detections to the frame.

    @param frame: np.ndarray frame
    @param dets: dai.ImgDetections
    @param norm: Object that handles normalization of the bounding box
    @param label_map: Label map for the detections
    """
    img_detections = []
    if isinstance(packet, TrackerPacket):
        img_detections = [t.srcImgDetection for t in packet.daiTracklets.tracklets]
    elif isinstance(packet, DetectionPacket):
        img_detections = [det for det in packet.img_detections.detections]

    for detection in img_detections:
        bbox = norm.get_relative_bbox(BoundingBox(detection)).to_tuple(packet.frame.shape)

        if label_map:
            txt, color = label_map[detection.label]
        else:
            txt = str(detection.label)
            color = VisualizerHelper.front_color

        VisualizerHelper.putText(packet.frame, txt, (bbox[0] + 5, bbox[1] + 25), scale=0.9)
        if packet._is_spatial_detection():
            point = packet._get_spatials(detection) \
                if isinstance(packet, TrackerPacket) \
                else detection.spatialCoordinates
            VisualizerHelper.putText(packet.frame, spatials_text(point).x, (bbox[0] + 5, bbox[1] + 50), scale=0.7)
            VisualizerHelper.putText(packet.frame, spatials_text(point).y, (bbox[0] + 5, bbox[1] + 75), scale=0.7)
            VisualizerHelper.putText(packet.frame, spatials_text(point).z, (bbox[0] + 5, bbox[1] + 100), scale=0.7)

        rectangle(packet.frame, bbox, color=color, thickness=1, radius=0)
        packet._add_detection(detection, bbox, txt, color)


def draw_tracklet_id(packet: TrackerPacket):
    for det in packet.detections:
        # centroid = det.centroid()
        VisualizerHelper.print_on_roi(packet.frame, det.top_left, det.bottom_right,
                                      f"Id: {str(det.tracklet.id)}",
                                      FramePosition.TopMid)


def draw_breadcrumb_trail(packets: List[TrackerPacket]):
    packet = packets[-1]  # Current packet

    dict_: Dict[str, List[TrackingDetection]] = {}
    valid_ids = [t.id for t in packet.daiTracklets.tracklets]
    for idx in valid_ids:
        dict_[str(idx)] = []

    for packet in packets:
        for det in packet.detections:
            if det.tracklet.id in valid_ids:
                dict_[str(det.tracklet.id)].append(det)

    for idx, list_ in dict_.items():
        for i in range(len(list_) - 1):
            VisualizerHelper.line(packet.frame, list_[i].centroid(), list_[i + 1].centroid(), color=list_[i].color)


def colorize_depth(depth_frame: Union[dai.ImgFrame, Any], color_map=None):
    if isinstance(depth_frame, dai.ImgFrame):
        depth_frame = depth_frame.getFrame()

    depth_frame_color = cv2.normalize(depth_frame, None, 256, 0, cv2.NORM_INF, cv2.CV_8UC3)
    depth_frame_color = cv2.equalizeHist(depth_frame_color)

    return cv2.applyColorMap(depth_frame_color, color_map or default_color_map)


def colorize_disparity(frame: Union[dai.ImgFrame, Any], multiplier: float, color_map=None):
    if isinstance(frame, dai.ImgFrame):
        frame = frame.getFrame()

    frame = (frame * multiplier).astype(np.uint8)

    return cv2.applyColorMap(frame, color_map or default_color_map)


def draw_bb_mappings(depth_frame: Union[dai.ImgFrame, Any], bb_mappings: dai.SpatialLocationCalculatorConfig):
    depth_frame_color = colorize_depth(depth_frame)
    roi_datas = bb_mappings.getConfigData()
    for roi_data in roi_datas:
        roi = roi_data.roi
        roi = roi.denormalize(depth_frame_color.shape[1], depth_frame_color.shape[0])
        top_left = roi.topLeft()
        bottom_right = roi.bottomRight()
        xmin = int(top_left.x)
        ymin = int(top_left.y)
        xmax = int(bottom_right.x)
        ymax = int(bottom_right.y)

        rectangle(depth_frame_color, (xmin, ymin, xmax, ymax), (255, 255, 255), 1)


def depth_to_disp_factor(device: dai.Device, stereo: dai.node.StereoDepth) -> float:
    """
    Calculates the disparity factor used to calculate disparity from depth, which is used for visualization.
    `disparity[0..95] = disparity_factor / depth`. We can then multiply disparity by 255/95 to get 0..255 range.
    @param device: OAK device
    """
    calib = device.readCalibration()
    cam1 = calib.getStereoLeftCameraId()
    cam2 = calib.getStereoRightCameraId()
    baseline = calib.getBaselineDistance(cam1=cam1, cam2=cam2, useSpecTranslation=True) * 10  # cm to mm
    raw_conf = stereo.initialConfig.get()

    align: dai.CameraBoardSocket = stereo.properties.depthAlignCamera
    if align == dai.CameraBoardSocket.AUTO:
        align = cam2

    intrinsics = calib.getCameraIntrinsics(align)
    focal_length = intrinsics[0][0]

    factor = baseline * focal_length
    if raw_conf.algorithmControl.enableExtended:
        factor /= 2

    return factor


def draw_bbox(img: np.ndarray,
              pt1: Tuple[int, int],
              pt2: Tuple[int, int],
              color: Tuple[int, int, int],
              thickness: int,
              r: int,
              line_width: int,
              line_height: int,
              alpha: float
              ) -> None:
    """
    Draw a rounded rectangle on the image (in-place).

    Args:
        img: Image to draw on.
        pt1: Top-left corner of the rectangle.
        pt2: Bottom-right corner of the rectangle.
        color: Rectangle color.
        thickness: Rectangle line thickness.
        r: Radius of the rounded corners.
        line_width: Width of the rectangle line.
        line_height: Height of the rectangle line.
        alpha: Opacity of the rectangle.
    """
    x1, y1 = pt1
    x2, y2 = pt2

    if line_width == 0:
        line_width = np.abs(x2 - x1)
        line_width -= 2 * r if r > 0 else 0  # Adjust for rounded corners

    if line_height == 0:
        line_height = np.abs(y2 - y1)
        line_height -= 2 * r if r > 0 else 0  # Adjust for rounded corners

    # Top left
    cv2.line(img, (x1 + r, y1), (x1 + r + line_width, y1), color, thickness)
    cv2.line(img, (x1, y1 + r), (x1, y1 + r + line_height), color, thickness)
    cv2.ellipse(img, (x1 + r, y1 + r), (r, r), 180, 0, 90, color, thickness)

    # Top right
    cv2.line(img, (x2 - r, y1), (x2 - r - line_width, y1), color, thickness)
    cv2.line(img, (x2, y1 + r), (x2, y1 + r + line_height), color, thickness)
    cv2.ellipse(img, (x2 - r, y1 + r), (r, r), 270, 0, 90, color, thickness)

    # Bottom left
    cv2.line(img, (x1 + r, y2), (x1 + r + line_width, y2), color, thickness)
    cv2.line(img, (x1, y2 - r), (x1, y2 - r - line_height), color, thickness)
    cv2.ellipse(img, (x1 + r, y2 - r), (r, r), 90, 0, 90, color, thickness)

    # Bottom right
    cv2.line(img, (x2 - r, y2), (x2 - r - line_width, y2), color, thickness)
    cv2.line(img, (x2, y2 - r), (x2, y2 - r - line_height), color, thickness)
    cv2.ellipse(img, (x2 - r, y2 - r), (r, r), 0, 0, 90, color, thickness)

    # Fill the area
    if 0 < alpha:
        overlay = img.copy()

        thickness = -1
        bbox = (pt1[0], pt1[1], pt2[0], pt2[1])

        top_left = (bbox[0], bbox[1])
        bottom_right = (bbox[2], bbox[3])
        top_right = (bottom_right[0], top_left[1])
        bottom_left = (top_left[0], bottom_right[1])

        top_left_main_rect = (int(top_left[0] + r), int(top_left[1]))
        bottom_right_main_rect = (int(bottom_right[0] - r), int(bottom_right[1]))

        top_left_rect_left = (top_left[0], top_left[1] + r)
        bottom_right_rect_left = (bottom_left[0] + r, bottom_left[1] - r)

        top_left_rect_right = (top_right[0] - r, top_right[1] + r)
        bottom_right_rect_right = (bottom_right[0], bottom_right[1] - r)

        all_rects = [
            [top_left_main_rect, bottom_right_main_rect],
            [top_left_rect_left, bottom_right_rect_left],
            [top_left_rect_right, bottom_right_rect_right]
        ]

        [cv2.rectangle(overlay, pt1=rect[0], pt2=rect[1], color=color, thickness=thickness) for rect in all_rects]

        cv2.ellipse(overlay, (top_left[0] + r, top_left[1] + r), (r, r), 180.0, 0, 90, color, thickness)
        cv2.ellipse(overlay, (top_right[0] - r, top_right[1] + r), (r, r), 270.0, 0, 90, color, thickness)
        cv2.ellipse(overlay, (bottom_right[0] - r, bottom_right[1] - r), (r, r), 0.0, 0, 90, color, thickness)
        cv2.ellipse(overlay, (bottom_left[0] + r, bottom_left[1] - r), (r, r), 90.0, 0, 90, color, thickness)

        cv2.addWeighted(overlay, alpha, img, 1 - alpha, 0, img)


def draw_stylized_bbox(img: np.ndarray, obj: VisBoundingBox) -> None:
    """
    Draw a stylized bounding box. The style is either passed as an argument or defined in the config.

    Args:
        img: Image to draw on.
        obj: Bounding box to draw.
    """
    pt1, pt2 = obj.bbox.denormalize(img.shape)

    box_w = pt2[0] - pt1[0]
    box_h = pt2[1] - pt1[1]

    line_width = int(box_w * obj.config.detection.line_width) // 2
    line_height = int(box_h * obj.config.detection.line_height) // 2
    roundness = int(obj.config.detection.box_roundness)
    bbox_style = obj.bbox_style or obj.config.detection.bbox_style
    alpha = obj.config.detection.fill_transparency

    if bbox_style == BboxStyle.RECTANGLE:
        draw_bbox(img, pt1, pt2,
                  obj.color, obj.thickness, 0,
                  line_width=0, line_height=0, alpha=alpha)
    elif bbox_style == BboxStyle.CORNERS:
        draw_bbox(img, pt1, pt2,
                  obj.color, obj.thickness, 0,
                  line_width=line_width, line_height=line_height, alpha=alpha)
    elif bbox_style == BboxStyle.ROUNDED_RECTANGLE:
        draw_bbox(img, pt1, pt2,
                  obj.color, obj.thickness, roundness,
                  line_width=0, line_height=0, alpha=alpha)
    elif bbox_style == BboxStyle.ROUNDED_CORNERS:
        draw_bbox(img, pt1, pt2,
                  obj.color, obj.thickness, roundness,
                  line_width=line_width, line_height=line_height, alpha=alpha)



import colorsys
import math
from typing import Tuple


def generate_colors(number_of_colors: int, pastel=0.5):
    colors = []

    # Calculate the number of hues and value steps
    steps = math.ceil(math.sqrt(number_of_colors))

    for i in range(steps):
        hue = i / steps

        for j in range(steps):
            value = 0.6 + (j / steps) * 0.4  # This will give a value between 0.6 and 1

            r, g, b = colorsys.hsv_to_rgb(hue, pastel, value)
            r, g, b = int(r * 255), int(g * 255), int(b * 255)
            colors.append((r, g, b))

    # Randomize colors
    # np.random.shuffle(colors)

    # Return only the first `number_of_colors` colors
    return colors[:number_of_colors]


def get_text_color(background_color: Tuple[int, int, int], threshold=0.6):
    """
    Determines whether black or white text will be more legible against a given background color.

    Args:
        background_color: The BGR color that the text will be displayed on.
        threshold: Float between 0 and 1. A threshold closer to 1 results in the function choosing white text more often.

    Returns:
        (0,0,0) for black text or (255,255,255) for white text.
    """
    blue, green, red = background_color
    if (red * 0.299 + green * 0.587 + blue * 0.114) > threshold:
        return (0, 0, 0)  # BGR for black
    else:
        return (255, 255, 255)  # BGR for white


def hex_to_bgr(hex: str) -> Tuple[int, ...]:
    """
    "#ff1f00" (red) => (0, 31, 255)
    """
    value = hex.lstrip('#')
    return tuple(int(value[i:i + 2], 16) for i in (4, 2, 0))



import numpy as np


def sharpness(frame: np.ndarray):
    dx = np.diff(frame)[1:, :]  # remove the first row
    dy = np.diff(frame, axis=0)[:, 1:]  # remove the first column
    dnorm = np.sqrt(dx ** 2 + dy ** 2)
    sharpness = np.average(dnorm)
    return sharpness



import json
import warnings
from collections import defaultdict
from pathlib import Path
from typing import Callable, Union, List, Dict

from depthai_sdk.types import NNNode
from depthai_sdk.visualize.bbox import BoundingBox

try:
    import blobconverter
except ImportError:
    blobconverter = None

from depthai_sdk.classes.nn_config import Config
from depthai_sdk.components.camera_component import CameraComponent
from depthai_sdk.components.component import Component, ComponentOutput
from depthai_sdk.integrations.roboflow import RoboflowIntegration
from depthai_sdk.components.multi_stage_nn import MultiStageNN
from depthai_sdk.components.nn_helper import *
from depthai_sdk.classes.enum import ResizeMode
from depthai_sdk.components.parser import *
from depthai_sdk.components.stereo_component import StereoComponent
from depthai_sdk.logger import LOGGER
from depthai_sdk.visualize.visualizer_helper import depth_to_disp_factor
from depthai_sdk.oak_outputs.xout.xout_base import StreamXout, XoutBase
from depthai_sdk.oak_outputs.xout.xout_frames import XoutFrames
from depthai_sdk.oak_outputs.xout.xout_nn import XoutTwoStage, XoutNnResults, XoutSpatialBbMappings, XoutNnData
from depthai_sdk.oak_outputs.xout.xout_tracker import XoutTracker
from depthai_sdk.replay import Replay


class NNComponent(Component):
    def __init__(self,
                 device: dai.Device,
                 pipeline: dai.Pipeline,
                 model: Union[str, Path, Dict],  # str for SDK supported model or Path to custom model's json
                 input: Union[CameraComponent, 'NNComponent', dai.Node.Output],
                 nn_type: Optional[str] = None,  # Either 'yolo' or 'mobilenet'
                 decode_fn: Optional[Callable] = None,
                 tracker: bool = False,  # Enable object tracker - only for Object detection models
                 spatial: Optional[StereoComponent] = None,
                 replay: Optional[Replay] = None,
                 args: Dict = None,  # User defined args
                 ) -> None:
        """
        Neural Network component abstracts:
         - DepthAI API nodes: NeuralNetwork, *DetectionNetwork, *SpatialDetectionNetwork, ObjectTracker
         - Downloading NN models (supported SDK NNs), parsing NN json configs and setting up the pipeline based on it
         - Decoding NN results
         - MultiStage pipelines - cropping high-res frames based on detections and use them for second NN inferencing

        Args:
            model (Union[str, Path, Dict]): str for SDK supported model / Path to blob or custom model's json
            input: (Union[Component, dai.Node.Output, dai.Node.Output]): Input to the NN. If nn_component that is object detector, crop HQ frame at detections (Script node + ImageManip node)
            nn_type (str, optional): Type of the NN - Either 'Yolo' or 'MobileNet'
            tracker (bool, default False): Enable object tracker - only for Object detection models
            spatial (bool, default False): Enable getting Spatial coordinates (XYZ), only for Obj detectors. Yolo/SSD use on-device spatial calc, others on-host (gen2-calc-spatials-on-host)
            replay (Replay object): Replay
            args (Any, optional): Use user defined arguments when constructing the pipeline
        """
        super().__init__()

        self.out = self.Out(self)

        self.triggers = defaultdict(list)
        self.node: Optional[NNNode] = None

        # ImageManip used to resize the input to match the expected NN input size
        self.image_manip: Optional[dai.node.ImageManip] = None
        self.x_in: Optional[dai.node.XLinkIn] = None  # Used for multi-stage pipeline
        # Tracker:
        self.tracker = pipeline.createObjectTracker() if tracker else None
        self.apply_tracking_filter = True  # Enable by default
        self.calculate_speed = True
        self.forget_after_n_frames = None

        # Private properties
        self._ar_resize_mode: ResizeMode = ResizeMode.LETTERBOX  # Default
        # Input to the NNComponent node passed on initialization
        self._input: Union[CameraComponent, 'NNComponent', dai.Node.Output] = input
        self._stream_input: dai.Node.Output  # Node Output that will be used as the input for this NNComponent

        self._blob: Optional[dai.OpenVINO.Blob] = None
        self._forced_version: Optional[dai.OpenVINO.Version] = None  # Forced OpenVINO version
        self._size: Optional[Tuple[int, int]] = None  # Input size to the NN
        self._args: Optional[Dict] = args
        self._config: Optional[Dict] = None
        self._node_type: dai.node = dai.node.NeuralNetwork  # Type of the node for `node`
        self._roboflow: Optional[RoboflowIntegration] = None

        # Multi-stage pipeline
        self._multi_stage_nn: Optional[MultiStageNN] = None

        self._input_queue = Optional[None]  # Input queue for multi-stage pipeline

        self._spatial: Optional[StereoComponent] = spatial
        self._replay: Optional[Replay] = replay  # Replay module

        # For visualizer
        self._labels: Optional[List] = None  # Obj detector labels
        self._handler: Optional[Callable] = None  # Custom model handler for decoding

        # Save passed settings
        self._decode_fn = decode_fn or None  # Decode function that will be used to decode NN results

        # Parse passed settings
        self._parse_model(model)
        if nn_type:
            self._parse_node_type(nn_type)

        # Create NN node
        self.node = pipeline.create(self._node_type)
        self._update_config()

        if self._roboflow:
            path = self._roboflow.device_update(device)
            self._parse_config(path)
            self._update_config()

        if self._blob is None:
            self._blob = dai.OpenVINO.Blob(self._blob_from_config(
                self._config['model'],
                self._config.get('openvino_version', None)
            ))

        # TODO: update NN input based on camera resolution
        self.node.setBlob(self._blob)
        self._out = self.node.out

        if 1 < len(self._blob.networkInputs):
            raise NotImplementedError()

        nn_in: dai.TensorInfo = next(iter(self._blob.networkInputs.values()))
        # TODO: support models that expect mono img
        self._size: Tuple[int, int] = (nn_in.dims[0], nn_in.dims[1])

        # Creates ImageManip node that resizes the input to match the expected NN input size.
        # DepthAI uses CHW (Planar) channel layout and BGR color order convention.
        self.image_manip = pipeline.createImageManip()
        # Configures ImageManip node. Letterbox by default
        self._change_resize_mode(ResizeMode.LETTERBOX)

        if isinstance(self._input, CameraComponent):
            self._stream_input = self._input.stream
            self._stream_input.link(self.image_manip.inputImage)
            # Link ImageManip output to NN node
            self.image_manip.out.link(self.node.input)
        elif isinstance(self._input, dai.Node.Output):
            self._stream_input = self._input
            self._stream_input.link(self.image_manip.inputImage)
            # Link ImageManip output to NN node
            self.image_manip.out.link(self.node.input)
        elif self.is_multi_stage():
            # Here, ImageManip will only crop the high-res frame to correct aspect ratio
            # (without resizing!) and it also acts as a buffer (by default, its pool size is set to 20).
            self.image_manip = pipeline.createImageManip()
            self.image_manip.setNumFramesPool(20)
            self._input._stream_input.link(self.image_manip.inputImage)
            frame_full_size = self._get_input_frame_size()

            if self._input.is_detector():
                self.image_manip.setMaxOutputFrameSize(frame_full_size[0] * frame_full_size[1] * 3)

                # Create script node, get HQ frames from input.
                self._multi_stage_nn = MultiStageNN(pipeline=pipeline,
                                                    detection_node=self._input.node,
                                                    high_res_frames=self.image_manip.out,
                                                    size=self._size,
                                                    frame_size=frame_full_size,
                                                    det_nn_size=self._input._size,
                                                    resize_mode=self._input._ar_resize_mode,
                                                    num_frames_pool=20)
                self._multi_stage_nn.out.link(self.node.input)  # Cropped frames

                # For debugging, for integral counter
                self.node.out.link(self._multi_stage_nn.script.inputs['recognition'])
                self.node.input.setBlocking(True)
                self.node.input.setQueueSize(20)
            else:
                LOGGER.debug('Using on-host decoding for multi-stage NN')
                # Custom NN
                self.image_manip.setResize(*self._size)
                self.image_manip.setMaxOutputFrameSize(self._size[0] * self._size[1] * 3)

                # TODO pass frame on device, and just send config from host
                self.x_in = pipeline.createXLinkIn()
                self.x_in.setStreamName("input_queue")
                self.x_in.setMaxDataSize(frame_full_size[0] * frame_full_size[1] * 3)
                self.x_in.out.link(self.image_manip.inputImage)

                self.x_in_cfg = pipeline.createXLinkIn()
                self.x_in_cfg.setStreamName("input_queue_cfg")
                self.x_in_cfg.out.link(self.image_manip.inputConfig)

                self.image_manip.out.link(self.node.input)
                self.node.input.setQueueSize(20)
        else:
            raise ValueError(
                "'input' argument passed on init isn't supported!"
                "You can only use NnComponent or CameraComponent as the input."
            )

        if self._spatial:
            self._stereo_node: dai.node.StereoDepth = self._spatial.node
            self._spatial.depth.link(self.node.inputDepth)
            self._spatial.config_stereo(align=self._input)
            # Configure Spatial Detection Network

        if self._args:
            if self.is_spatial():
                self._config_spatials_args(self._args)

    def get_name(self):
        model = self._config.get('model', None)
        if model is not None:
            return model.get('model_name', None)
        return None

    def get_labels(self):
        return [l.upper() if isinstance(l, str) else l[0].upper() for l in self._labels]

    def _parse_model(self, model):
        """
        Called when NNComponent is initialized. Parses "model" argument passed by user.
        """
        if isinstance(model, Dict):
            self._parse_config(model)
            return
        # Parse the input config/model
        elif isinstance(model, str):
            # Download from the web, or convert to Path
            model = getBlob(model) if isUrl(model) else Path(model)

        if model.suffix in ['.blob', '.json']:
            if model.suffix == '.blob':
                self._blob = dai.OpenVINO.Blob(model.resolve())
                self._forced_version = self._blob.version
            elif model.suffix == '.json':  # json config file was passed
                self._parse_config(model)
        else:  # SDK supported model
            models = getSupportedModels(printModels=False)
            zoo_models = blobconverter.zoo_list()

            if str(model) in models:
                model = models[str(model)] / 'config.json'
                self._parse_config(model)
            elif str(model) in zoo_models:
                LOGGER.warning(
                    'Models from the OpenVINO Model Zoo do not carry any metadata'
                    ' (e.g., label map, decoding logic). Please keep this in mind when using models from Zoo.'
                )
                self._blob = dai.OpenVINO.Blob(blobconverter.from_zoo(str(model), shaves=6))
                self._forced_version = self._blob.version
            else:
                raise ValueError(f"Specified model '{str(model)}' is not supported by DepthAI SDK.\n"
                                 "Check SDK documentation page to see which models are supported.")

    def _parse_node_type(self, nn_type: str) -> None:
        self._node_type = dai.node.NeuralNetwork
        if nn_type:
            if nn_type.upper() == 'YOLO':
                self._node_type = dai.node.YoloSpatialDetectionNetwork if self.is_spatial() else dai.node.YoloDetectionNetwork
            elif nn_type.upper() == 'MOBILENET':
                self._node_type = dai.node.MobileNetSpatialDetectionNetwork if self.is_spatial() else dai.node.MobileNetDetectionNetwork

    def _config_spatials_args(self, args):
        if not isinstance(args, Dict):
            args = vars(args)  # Namespace -> Dict
        self.config_spatial(
            bb_scale_factor=args.get('sbbScaleFactor', None),
            lower_threshold=args.get('minDepth', None),
            upper_threshold=args.get('maxDepth', None),
        )

    def _parse_config(self, model_config: Union[Path, str, Dict]):
        """
        Called when NNComponent is initialized. Reads config.json file and parses relevant setting from there
        """
        parent_folder = None
        if isinstance(model_config, str):
            model_config = Path(model_config).resolve()

        if isinstance(model_config, Path):
            parent_folder = model_config.parent
            with model_config.open() as f:
                self._config = Config().load(json.loads(f.read()))
        else:  # Dict
            self._config = model_config

        if 'source' in self._config:
            if self._config['source'] == 'roboflow':
                from depthai_sdk.integrations.roboflow import RoboflowIntegration
                self._roboflow = RoboflowIntegration(self._config)
                self._parse_node_type('YOLO')  # Roboflow only supports YOLO models
                return
            else:
                raise ValueError(f"[NN Dict configuration] Source '{self._config['source']}' not supported")

        # Get blob from the config file
        if 'model' in self._config:
            model = self._config['model']

            # Resolve the paths inside config
            if parent_folder:
                for name in ['blob', 'xml', 'bin']:
                    if name in model:
                        model[name] = str((parent_folder / model[name]).resolve())

            if 'blob' in model:
                self._blob = dai.OpenVINO.Blob(model['blob'])

        # Parse OpenVINO version
        if "openvino_version" in self._config:
            self._forced_version = parse_open_vino_version(self._config.get("openvino_version"))

        # Save for visualization
        self._labels = self._config.get("mappings", {}).get("labels", None)

        # Handler.py logic to decode raw NN results into standardized AI results
        if 'handler' in self._config:
            self._handler = loadModule(model_config.parent / self._config["handler"])

            if not callable(getattr(self._handler, "decode", None)):
                LOGGER.debug("Custom model handler does not contain 'decode' method!")
            else:
                self._decode_fn = self._handler.decode if self._decode_fn is None else self._decode_fn

        if 'nn_config' in self._config:
            nn_config = self._config.get("nn_config", {})

            # Parse node type
            nn_family = nn_config.get("NN_family", None)
            if nn_family:
                self._parse_node_type(nn_family)

    def _blob_from_config(self, model: Dict, version: Union[None, str, dai.OpenVINO.Version] = None) -> str:
        """
        Gets the blob from the config file.
        """
        if isinstance(version, dai.OpenVINO.Version):
            version = str(version)
        if isinstance(version, str):
            if version.startswith('VERSION_'):
                version = version[8:]
            if '_' in version:
                vals = version.split('_')
                version = f'{vals[0]}.{vals[1]}'

        if 'model_name' in model:  # Use blobconverter to download the model
            zoo_type = model.get("zoo", 'intel')
            return blobconverter.from_zoo(model['model_name'],
                                          zoo_type=zoo_type,
                                          shaves=6,  # TODO: Calculate ideal shave amount
                                          version=version)

        if 'xml' in model and 'bin' in model:
            return blobconverter.from_openvino(xml=model['xml'],
                                               bin=model['bin'],
                                               data_type="FP16",  # Myriad X
                                               shaves=6,  # TODO: Calculate ideal shave amount
                                               version=version)

        raise ValueError("Specified `model` values in json config files are incorrect!")

    def _change_resize_mode(self, mode: ResizeMode) -> None:
        """
        Changes the resize mode of the ImageManip node.

        Args:
            mode (ResizeMode): Resize mode to use
        """
        if self.is_multi_stage():
            return  # We need high-res frames for multi-stage NN, so we can crop them later

        self._ar_resize_mode = mode

        # Reset ImageManip node config
        self.image_manip.initialConfig.set(dai.RawImageManipConfig())
        self.image_manip.setFrameType(dai.RawImgFrame.Type.BGR888p)
        self.image_manip.setMaxOutputFrameSize(self._size[0] * self._size[1] * 3)
        self.image_manip.inputImage.setBlocking(False)
        self.image_manip.inputImage.setQueueSize(2)

        if self._ar_resize_mode == ResizeMode.CROP:
            self.image_manip.initialConfig.setResize(self._size)
            # With .setCenterCrop(1), ImageManip will first crop the frame to the correct aspect ratio,
            # and then resize it to the NN input size
            self.image_manip.initialConfig.setCenterCrop(1, self._size[0] / self._size[1])
        elif self._ar_resize_mode == ResizeMode.LETTERBOX:
            self.image_manip.initialConfig.setResizeThumbnail(*self._size)
        elif self._ar_resize_mode == ResizeMode.STRETCH:
            # Not keeping aspect ratio -> stretching the image
            self.image_manip.initialConfig.setKeepAspectRatio(False)
            self.image_manip.initialConfig.setResize(self._size)

    def config_multistage_nn(self,
                             debug=False,
                             labels: Optional[List[int]] = None,
                             scale_bb: Optional[Tuple[int, int]] = None,
                             num_frame_pool: int = None
                             ) -> None:
        """
        Configures the MultiStage NN pipeline. Available if the input to this NNComponent is Detection NNComponent.

        Args:
            debug (bool, default False): Debug script node
            labels (List[int], optional): Crop & run inference only on objects with these labels
            scale_bb (Tuple[int, int], optional): Scale detection bounding boxes (x, y) before cropping the frame. In %.
            num_frame_pool (int, optional): Number of frames to pool for inference. If None, will use the default value.
        """
        if not self.is_multi_stage():
            LOGGER.warning("Input to this model was not a NNComponent, so 2-stage NN inferencing isn't possible!"
                            "This configuration attempt will be ignored.")
            return

        if num_frame_pool is not None:
            # This ImageManip crops the the saved frame based on the detection.
            # Script node sends frame + config to it dynamically.
            self._multi_stage_nn.manip.setNumFramesPool(num_frame_pool)
            # This ImageManip crops the high-res frames and acts as a before
            # prior to Script node (which saves these cropped frames in an array)
            self.image_manip.setNumFramesPool(num_frame_pool)

        self._multi_stage_nn.configure(debug, labels, scale_bb)

    def _parse_label(self, label: Union[str, int]) -> int:
        if isinstance(label, int):
            return label
        elif isinstance(label, str):
            if not self._labels:
                raise ValueError("Incorrect trackLabels type! Make sure to pass NN configuration to "
                                 "the NNComponent so it can deccode string labels!")
            # Label map is Dict of either "name", or ["name", "color"]
            label_strs = [l.upper() if isinstance(l, str) else l[0].upper() for l in self._labels]

            if label.upper() not in label_strs: raise ValueError(f"String '{label}' wasn't found in passed labels!")
            return label_strs.index(label.upper())
        else:
            raise Exception('_parse_label only accepts int or str')

    def config_tracker(self,
                       tracker_type: Optional[dai.TrackerType] = None,
                       track_labels: Optional[List[Union[int, str]]] = None,
                       assignment_policy: Optional[dai.TrackerIdAssignmentPolicy] = None,
                       max_obj: Optional[int] = None,
                       threshold: Optional[float] = None,
                       apply_tracking_filter: Optional[bool] = None,
                       forget_after_n_frames: Optional[int] = None,
                       calculate_speed: Optional[bool] = None
                       ) -> None:
        """
        Configure Object Tracker node (if it's enabled).

        Args:
            tracker_type (dai.TrackerType, optional): Set object tracker type
            track_labels (List[int], optional): Set detection labels to track
            assignment_policy (dai.TrackerType, optional): Set object tracker ID assignment policy
            max_obj (int, optional): Set max objects to track. Max 60.
            threshold (float, optional): Specify tracker threshold. Default: 0.0
            apply_tracking_filter (bool, optional): Set whether to apply Kalman filter to the tracked objects. Done on the host.
            forget_after_n_frames (int, optional): Set how many frames to track an object before forgetting it.
            calculate_speed (bool, optional): Set whether to calculate object speed. Done on the host.
        """

        if self.tracker is None:
            warnings.warn("Tracker was not enabled! Enable with cam.create_nn('[model]', tracker=True)."
                          "This configuration attempt will be ignored.")
            return

        if tracker_type is not None:
            self.tracker.setTrackerType(type=tracker_type)

        if track_labels is not None and 0 < len(track_labels):
            labels = [self._parse_label(l) for l in track_labels]
            self.tracker.setDetectionLabelsToTrack(labels)

        if assignment_policy is not None:
            self.tracker.setTrackerIdAssignmentPolicy(assignment_policy)

        if max_obj is not None:
            if 60 < max_obj:
                raise ValueError("Maximum objects to track is 60!")
            self.tracker.setMaxObjectsToTrack(max_obj)

        if threshold is not None:
            self.tracker.setTrackerThreshold(threshold)

        if apply_tracking_filter is not None:
            self.apply_tracking_filter = apply_tracking_filter

        if forget_after_n_frames is not None:
            self.forget_after_n_frames = forget_after_n_frames

        if calculate_speed is not None:
            self.calculate_speed = calculate_speed

    def config_yolo_from_metadata(self, metadata: Dict) -> None:
        """
        Configures (Spatial) Yolo Detection Network node with a dictionary. Calls config_yolo().
        """
        self.config_yolo(
            num_classes=metadata['classes'],
            coordinate_size=metadata['coordinates'],
            anchors=metadata['anchors'],
            masks=metadata['anchor_masks'],
            iou_threshold=metadata['iou_threshold'],
            conf_threshold=metadata['confidence_threshold'],
        )

    def config_yolo(self,
                    num_classes: int,
                    coordinate_size: int,
                    anchors: List[float],
                    masks: Dict[str, List[int]],
                    iou_threshold: float,
                    conf_threshold: Optional[float] = None,
                    ) -> None:
        """
        Configures (Spatial) Yolo Detection Network node.
        """
        if not self.is_yolo():
            LOGGER.warning('This is not a YOLO detection network! This configuration attempt will be ignored.')
            return

        if not self.node:
            raise Exception('YOLO node not initialized!')

        self.node.setNumClasses(num_classes)
        self.node.setCoordinateSize(coordinate_size)
        self.node.setAnchors(anchors)
        self.node.setAnchorMasks(masks)
        self.node.setIouThreshold(iou_threshold)

        if conf_threshold is not None:
            self.node.setConfidenceThreshold(conf_threshold)

    def config_nn(self,
                  conf_threshold: Optional[float] = None,
                  resize_mode: Union[ResizeMode, str] = None
                  ) -> None:
        """
        Configures the Detection Network node.

        Args:
            conf_threshold: (float, optional): Confidence threshold for the detections (0..1]
            resize_mode: (ResizeMode, optional): Change aspect ratio resizing mode - to either STRETCH, CROP, or LETTERBOX.
        """
        if resize_mode:
            self._ar_resize_mode = ResizeMode.parse(resize_mode)
            self._change_resize_mode(self._ar_resize_mode)

        if conf_threshold is not None and self.is_detector():
            if 0 <= conf_threshold <= 1:
                self.node.setConfidenceThreshold(conf_threshold)
            else:
                raise ValueError("Confidence threshold must be between 0 and 1!")

    def config_spatial(self,
                       bb_scale_factor: Optional[float] = None,
                       lower_threshold: Optional[int] = None,
                       upper_threshold: Optional[int] = None,
                       calc_algo: Optional[dai.SpatialLocationCalculatorAlgorithm] = None
                       ) -> None:
        """
        Configures the Spatial Detection Network node.

        Args:
            bb_scale_factor (float, optional): Specifies scale factor for detected bounding boxes (0..1]
            lower_threshold (int, optional): Specifies lower threshold in depth units (millimeter by default) for depth values which will used to calculate spatial data
            upper_threshold (int, optional): Specifies upper threshold in depth units (millimeter by default) for depth values which will used to calculate spatial data
            calc_algo (dai.SpatialLocationCalculatorAlgorithm, optional): Specifies spatial location calculator algorithm: Average/Min/Max
        """
        if not self.is_spatial():
            LOGGER.warning('This is not a Spatial Detection network! This configuration attempt will be ignored.')
            return

        if bb_scale_factor is not None:
            self.node.setBoundingBoxScaleFactor(bb_scale_factor)
        if lower_threshold is not None:
            self.node.setDepthLowerThreshold(lower_threshold)
        if upper_threshold is not None:
            self.node.setDepthUpperThreshold(upper_threshold)
        if calc_algo:
            self.node.setSpatialCalculationAlgorithm(calc_algo)

    def _update_config(self) -> None:
        if self.node is None or self._config is None:
            return

        nn_config = self._config.get("nn_config", {})

        meta = nn_config.get('NN_specific_metadata', None)
        if self.is_yolo() and meta:
            self.config_yolo_from_metadata(metadata=meta)

        self.config_nn(conf_threshold=nn_config.get('conf_threshold', None))

    def _get_camera_comp(self) -> CameraComponent:
        if self.is_multi_stage():
            return self._input._get_camera_comp()
        return self._input

    def _get_input_frame_size(self) -> Tuple[int, int]:
        # TODO: if user passes node output as the NN input (eg. examples/mixed/switch_between_models.py),
        # this function will fail
        return self._get_camera_comp().stream_size

    #
    def get_bbox(self) -> BoundingBox:
        if self.is_multi_stage():
            return self._input.get_bbox()
        else:
            try:
                stream_size = self._get_input_frame_size()
                old_ar = stream_size[0] / stream_size[1]
                new_ar = self._size[0] / self._size[1]
                return BoundingBox().resize_to_aspect_ratio(old_ar, new_ar, self._ar_resize_mode)
            except (AttributeError, ZeroDivisionError, ValueError):
                return BoundingBox()

    """
    Available outputs (to the host) of this component
    """

    class Out:

        class MainOut(ComponentOutput):
            """
            Default output. Streams NN results and high-res frames that were downscaled and used for inferencing.
            Produces DetectionPacket or TwoStagePacket (if it's 2. stage NNComponent).
            """

            def __call__(self, device: dai.Device, fourcc: Optional[str] = None) -> XoutBase:
                if self._comp.is_multi_stage():
                    det_nn_out = StreamXout(out=self._comp._input.node.out)
                    second_nn_out = StreamXout(out=self._comp.node.out)

                    return XoutTwoStage(det_nn=self._comp._input,
                                        second_nn=self._comp,
                                        frames=self._comp._input._input.get_stream_xout(),
                                        det_out=det_nn_out,
                                        second_nn_out=second_nn_out,
                                        device=device,
                                        input_queue_name="input_queue" if self._comp.x_in else None,
                                        bbox=self._comp.get_bbox()).set_fourcc(fourcc).set_comp_out(self)
                else:
                    # TODO: refactor. This is a bit hacky, as we want to support passing node output as the input
                    # to the NNComponent. In such case, we don't have access to VideoEnc (inside CameraComponent)
                    det_nn_out = StreamXout(out=self._comp.node.out)
                    input_stream = self._comp._stream_input
                    if fourcc is None:
                        frame_stream = StreamXout(out=input_stream)
                    else:
                        frame_stream = self._comp._get_camera_comp().get_stream_xout(fourcc)
                    return XoutNnResults(det_nn=self._comp,
                                         frames=frame_stream,
                                         nn_results=det_nn_out,
                                         bbox=self._comp.get_bbox()).set_fourcc(fourcc).set_comp_out(self)

        class PassThroughOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                """
                Default output. Streams NN results and passthrough frames (frames used for inferencing)
                Produces DetectionPacket or TwoStagePacket (if it's 2. stage NNComponent).
                """
                if self._comp.is_multi_stage():
                    return XoutTwoStage(det_nn=self._comp._input,
                                        second_nn=self._comp,
                                        frames=StreamXout(out=self._comp._input.node.passthrough),
                                        det_out=StreamXout(out=self._comp._input.node.out),
                                        second_nn_out=StreamXout(self._comp.node.out),
                                        device=device,
                                        input_queue_name="input_queue" if self._comp.x_in else None,
                                        bbox=self._comp.get_bbox()).set_comp_out(self)
                else:
                    return XoutNnResults(det_nn=self._comp,
                                         frames=StreamXout(out=self._comp.node.passthrough),
                                         nn_results=StreamXout(out=self._comp.node.out),
                                         bbox=BoundingBox()
                                         ).set_comp_out(self)

        class ImgManipOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                return XoutFrames(StreamXout(out=self._comp.image_manip.out)).set_comp_out(self)

        class InputOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutBase:
                return XoutFrames(StreamXout(out=self._comp._stream_input)).set_comp_out(self)

        class SpatialOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutSpatialBbMappings:
                """
                Streams depth and bounding box mappings (``SpatialDetectionNework.boundingBoxMapping``). Produces SpatialBbMappingPacket.
                """
                if not self._comp.is_spatial():
                    raise Exception('SDK tried to output spatial data (depth + bounding box mappings),'
                                    'but this is not a Spatial Detection network!')

                return XoutSpatialBbMappings(
                    device=device,
                    stereo=self._comp._stereo_node,
                    frames=StreamXout(out=self._comp.node.passthroughDepth),
                    configs=StreamXout(out=self._comp.node.out),
                    dispScaleFactor=depth_to_disp_factor(device, self._comp._stereo_node),
                    bbox=self._comp.get_bbox()
                ).set_comp_out(self)

        class TwoStageOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutFrames:
                """
                Streams 2. stage cropped frames to the host. Produces FramePacket.
                """
                if not self._comp.is_multi_stage():
                    raise Exception(
                        'SDK tried to output TwoStage crop frames, but this is not a Two-Stage NN component!')

                return XoutFrames(frames=StreamXout(out=self._comp._multi_stage_nn.manip.out)).set_comp_out(self)

        class TrackerOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutTracker:
                """
                Streams ObjectTracker tracklets and high-res frames that were downscaled and used for inferencing. Produces TrackerPacket.
                """
                if not self._comp.is_tracker():
                    raise Exception('Tracker was not enabled! Enable with cam.create_nn("[model]", tracker=True)!')

                self._comp.node.passthrough.link(self._comp.tracker.inputDetectionFrame)
                self._comp.node.out.link(self._comp.tracker.inputDetections)

                # TODO: add support for full frame tracking
                self._comp.node.passthrough.link(self._comp.tracker.inputTrackerFrame)

                return XoutTracker(det_nn=self._comp,
                                   frames=self._comp._input.get_stream_xout(),  # CameraComponent
                                   device=device,
                                   tracklets=StreamXout(self._comp.tracker.out),
                                   bbox=self._comp.get_bbox(),
                                   apply_kalman=self._comp.apply_tracking_filter,
                                   forget_after_n_frames=self._comp.forget_after_n_frames,
                                   calculate_speed=self._comp.calculate_speed,
                                   ).set_comp_out(self)

        class EncodedOut(MainOut):
            def __call__(self, device: dai.Device) -> XoutNnResults:
                """
                Streams NN results and encoded frames (frames used for inferencing)
                Produces DetectionPacket or TwoStagePacket (if it's 2. stage NNComponent).
                """
                # A bit hacky, maybe we can remove this alltogether
                return super().__call__(device, fourcc=self._comp._get_camera_comp().get_fourcc())

        class NnDataOut(ComponentOutput):
            def __call__(self, device: dai.Device) -> XoutNnData:
                node_output = self._comp.node.out if \
                    type(self._comp.node) == dai.node.NeuralNetwork else \
                    self._comp.node.outNetwork

                return XoutNnData(xout=StreamXout(node_output)).set_comp_out(self)

        def __init__(self, nn_component: 'NNComponent'):
            self.main = self.MainOut(nn_component)
            self.passthrough = self.PassThroughOut(nn_component)
            self.image_manip = self.ImgManipOut(nn_component)
            self.input = self.InputOut(nn_component)
            self.spatials = self.SpatialOut(nn_component)
            self.twostage_crops = self.TwoStageOut(nn_component)
            self.tracker = self.TrackerOut(nn_component)
            self.encoded = self.EncodedOut(nn_component)
            self.nn_data = self.NnDataOut(nn_component)

    # Checks
    def is_spatial(self) -> bool:
        return self._spatial is not None  # todo fix if spatial is bool and equals to False

    def is_tracker(self) -> bool:
        # Currently, only object detectors are supported
        return self.is_detector() and self.tracker is not None

    def is_yolo(self) -> bool:
        return (
                self._node_type == dai.node.YoloDetectionNetwork or
                self._node_type == dai.node.YoloSpatialDetectionNetwork
        )

    def is_mobile_net(self) -> bool:
        return (
                self._node_type == dai.node.MobileNetDetectionNetwork or
                self._node_type == dai.node.MobileNetSpatialDetectionNetwork
        )

    def is_detector(self) -> bool:
        """
        Currently these 2 object detectors are supported
        """
        return self.is_yolo() or self.is_mobile_net()

    def is_multi_stage(self):
        if not isinstance(self._input, type(self)):
            return False

        if not isinstance(self._input, Component):
            return False

        # if not self._input._is_detector():
        #     raise Exception('Only object detector models can be used as an input to the NNComponent!')

        return True
